# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""
The :py:obj:`floris` package contains :py:obj:`floris.utilities` module
and the packages which make up the FLORIS software. The
:py:obj:`floris.simulation` package contains the modules used to
complete a wake simulation for a given wind farm and turbine
configuration. The :py:obj:`floris.tools` package contains the modules
used to drive FLORIS simulations and perform studies in various areas
of research and analysis. 

All modules and package can be imported with

    >>> import floris

The ``__init__.py`` file enables the import of all modules in this
package so any additional modules should be included there.

Examples:
    >>> import floris
    
    >>> dir(floris)
    ['__builtins__', '__cached__', '__doc__', '__file__', '__loader__',
    '__name__', '__package__', '__path__', '__spec__', 'simulation',
    'tools', 'utilities']

    >>> dir(floris.utilities)
    ['Output', 'Vec3', '__builtins__', '__cached__', '__doc__',
    '__file__', '__loader__', '__name__', '__package__', '__spec__',
    'cosd', 'np', 'sind', 'tand', 'wrap_180', 'wrap_360']

    >>> dir(floris.simulation)
    ['Farm', 'Floris', 'FlowField', 'InputReader', 'Turbine',
    'TurbineMap', 'Wake', 'WakeCombination', 'WakeDeflection',
    'WakeVelocity', '__builtins__', '__cached__', '__doc__',
    '__file__', '__loader__', '__name__', '__package__', '__path__',
    '__spec__', 'farm', 'floris', 'flow_field', 'input_reader',
    'turbine', 'turbine_map', 'wake', 'wake_combination',
    'wake_deflection', 'wake_velocity']

    >>> dir(floris.tools)
    ['__builtins__', '__cached__', '__doc__', '__file__', '__loader__',
    '__name__', '__package__', '__path__', '__spec__', 'cut_plane',
    'energy_ratio', 'floris_utilities', 'flow_data',
    'layout_functions', 'optimization', 'plotting', 'power_rose',
    'rews', 'sowfa_utilities', 'visualization', 'wind_rose']
"""

from . import utilities
from . import simulation
from . import tools



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from . import wake_deflection
from . import wake_velocity
from . import wake_combination


class Wake():
    """
    Wake is a container class for the various wake model objects. In
    particular, Wake holds references to the velocity and deflection
    models as well as their parameters.
    """

    def __init__(self, instance_dictionary):
        """
        Init method for Wake objects.

        Args:
            instance_dictionary (dict): the input dictionary with the
            following key-value pairs:
                {
                    "description": str,
                    "properties": dict({
                        velocity_model: WakeVelocity
                        deflection_model: WakeDeflection
                        parameters: dict({
                            see WakeVelocity, WakeDeflection
                        })
                    }),
                }
        """

        self.description = instance_dictionary["description"]
        properties = instance_dictionary["properties"]
        parameters = properties["parameters"]

        self.velocity_models = {
            "jensen": wake_velocity.Jensen(parameters),
            "multizone": wake_velocity.MultiZone(parameters),
            "gauss": wake_velocity.Gauss(parameters),
            "curl": wake_velocity.Curl(parameters)
        }
        self._velocity_model = self.velocity_models[
            properties["velocity_model"]]

        self.deflection_models = {
            "jimenez": wake_deflection.Jimenez(parameters),
            "gauss": wake_deflection.Gauss(parameters),
            "curl": wake_deflection.Curl(parameters)
        }
        self._deflection_model = self.deflection_models[
            properties["deflection_model"]]

        self.combination_models = {
            "fls": wake_combination.FLS(),
            "sosfs": wake_combination.SOSFS()
        }
        self._combination_model = self.combination_models[
            properties["combination_model"]]

    # Getters & Setters
    @property
    def velocity_model(self):
        """
        Print or re-assign the velocity model. Recognized types:

         - jensen
         - multizone
         - gauss
         - curl
        """
        return self._velocity_model

    @velocity_model.setter
    def velocity_model(self, value):
        self._velocity_model = self.velocity_models[value]

    @property
    def deflection_model(self):
        """
        Print or re-assign the deflection model. Recognized types:

         - jimenez
         - gauss
         - curl
        """
        return self._deflection_model

    @deflection_model.setter
    def deflection_model(self, value):
        self._deflection_model = self.deflection_models[value]

    @property
    def combination_model(self):
        """
        Print or re-assign the combination model. Recognized types:

         - fls
         - sosfs
        """
        return self._combination_model

    @combination_model.setter
    def combination_model(self, value):
        self._combination_model = self.combination_models[value]

    @property
    def deflection_function(self):
        """
        Return the underlying function of the deflection model.
        """
        return self._deflection_model.function

    @property
    def velocity_function(self):
        """
        Return the underlying function of the velocity model.
        """
        return self._velocity_model.function

    @property
    def combination_function(self):
        """
        Return the underlying function of the combination model.
        """
        return self._combination_model.function



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import numpy as np
from ..utilities import Vec3, Output


class FlowData():
    """
    Generate a FlowData object to handle data I/O
    """

    #TODO handle none case, maybe defaul values apply like 0 origin and auto determine spacing and dimensions
    def __init__(self,
                 x,
                 y,
                 z,
                 u,
                 v,
                 w,
                 spacing=None,
                 dimensions=None,
                 origin=None):
        """
        Initialize FlowData object with coordinates, velocity fields,
        and meta data.

        Args:
            x (np.array): Cartesian coordinate data.
            y (np.array): Cartesian coordinate data.
            z (np.array): Cartesian coordinate data.
            u (np.array): x-component of velocity.
            v (np.array): y-component of velocity.
            w (np.array): z-component of velocity.
            spacing (float, optional): Spatial resolution.
                Defaults to None.
            dimensions (iterable, optional): named dimensions
                (e.g. x1, x2, x3). Defaults to None.
            origin (iterable, optional): Coordinates of origin.
                Defaults to None.
        """

        self.x = x
        self.y = y
        self.z = z
        self.u = u
        self.v = v
        self.w = w

        #TODO Make these VEC3?
        self.spacing = spacing
        self.dimensions = dimensions
        self.origin = origin

        # Technically resolution is a restating of above, but it is useful to have
        self.resolution = Vec3(len(np.unique(x)), len(np.unique(y)),
                               len(np.unique(z)))

    def save_as_vtk(self, filename):
        """
        Save FlowData Object to vtk

        Args:
            filename (str): Write-to path for vtk file
        """
        n_points = self.dimensions.x1 * self.dimensions.x2 * self.dimensions.x3
        vtk_file = Output(filename)
        vtk_file.write_line('# vtk DataFile Version 3.0')
        vtk_file.write_line('array.mean0D')
        vtk_file.write_line('ASCII')
        vtk_file.write_line('DATASET STRUCTURED_POINTS')
        vtk_file.write_line('DIMENSIONS {}'.format(self.dimensions))
        vtk_file.write_line('ORIGIN {}'.format(self.origin))
        vtk_file.write_line('SPACING {}'.format(self.spacing))
        vtk_file.write_line('POINT_DATA {}'.format(n_points))
        vtk_file.write_line('FIELD attributes 1')
        vtk_file.write_line('UAvg 3 {} float'.format(n_points))
        for u, v, w in zip(self.u, self.v, self.w):
            vtk_file.write_line('{}'.format(Vec3(u, v, w)))

    @staticmethod
    def crop(ff, x_bnds, y_bnds, z_bnds):
        """
        Crop FlowData object to within stated bounds.

        Args:
            ff (:py:class:`floris.tools.flow_data.FlowData`):
                FlowData object.
            x_bnds (iterable): min and max of x-coordinate.
            y_bnds (iterable): min and max of y-coordinate.
            z_bnds (iterable): min and max of z-coordinate.

        Returns:
            (:py:class:`floris.tools.flow_data.FlowData`):
            cropped FlowData object.
        """

        map_values = (ff.x > x_bnds[0]) & (ff.x < x_bnds[1]) & (
            ff.y > y_bnds[0]) & (ff.y < y_bnds[1]) & (ff.z > z_bnds[0]) & (
                ff.z < z_bnds[1])

        x = ff.x[map_values]
        y = ff.y[map_values]
        z = ff.z[map_values]

        
        dimensions = Vec3(len(np.unique(x)), len(np.unique(y)), len(np.unique(z)))

        # Work out origin
        origin = Vec3(
            ff.origin.x1 + np.min(x),
            ff.origin.x2 + np.min(y),
            ff.origin.x3 + np.min(z),
        )


        return FlowData(
            x - np.min(x),
            y - np.min(y),
            z - np.min(z),
            ff.u[map_values],
            ff.v[map_values],
            ff.w[map_values],
            spacing=ff.spacing,  # doesn't change
            dimensions=dimensions,
            origin=origin)



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np
from floris.simulation import Floris
from floris.simulation import TurbineMap
from .flow_data import FlowData
from ..utilities import Vec3
import copy


class FlorisInterface():
    """
    The interface between a FLORIS instance and the wfc tools
    """

    def __init__(self, input_file):
        self.input_file = input_file
        self.floris = Floris(input_file=input_file)

    def calculate_wake(self, yaw_angles=None):
        """
        Wrapper to the floris flow field calculate_wake method

        Args:
            yaw_angles (np.array, optional): Turbine yaw angles.
                Defaults to None.
        """

        if yaw_angles is not None:
            self.floris.farm.set_yaw_angles(yaw_angles)

        self.floris.farm.flow_field.calculate_wake()

    def reinitialize_flow_field(self,
                                wind_speed=None,
                                wind_direction=None,
                                wind_shear=None,
                                wind_veer=None,
                                turbulence_intensity=None,
                                air_density=None,
                                wake=None,
                                layout_array=None,
                                with_resolution=None):
        """
        Wrapper to
        :py:meth:`floris.simlulation.flow_field.reinitialize_flow_field`.
        All input values are used to update the flow_field instance.

        Args:
            wind_speed (float, optional): background wind speed.
                Defaults to None.
            wind_direction (float, optional): background wind direction.
                Defaults to None.
            wind_shear (float, optional): shear exponent.
                Defaults to None.
            wind_veer (float, optional): direction change over rotor.
                Defaults to None.
            turbulence_intensity (float, optional): background turbulence intensity. Defaults to None.Defaults to None.
            (float, optional): ambient air density.
                Defaults to None.
            wake (str, optional): wake model type. Defaults to None.
            layout_array (np.array, optional): array of x- and
                y-locations of wind turbines. Defaults to None.
            with_resolution (float, optional): resolution of output
                flow_field. Defaults to None.
        """

        # Build turbine map (convenience layer for user)
        if layout_array is not None:
            turbine_map = TurbineMap(
                layout_array[0], layout_array[1], \
                [copy.deepcopy(self.floris.farm.turbines[0]) \
                for ii in range(len(layout_array[0]))])
        else:
            turbine_map = None

        self.floris.farm.flow_field.reinitialize_flow_field(
            wind_speed=wind_speed,
            wind_direction=wind_direction,
            wind_shear=wind_shear,
            wind_veer=wind_veer,
            turbulence_intensity=turbulence_intensity,
            air_density=air_density,
            wake=wake,
            turbine_map=turbine_map,
            with_resolution=with_resolution)

    # Special case function for quick visualization of hub height
    def get_hub_height_flow_data(self,
                                 x_resolution=100,
                                 y_resolution=100,
                                 x_bounds=None,
                                 y_bounds=None):
        """
        Shortcut method to visualize flow field at hub height.

        Args:
            x_resolution (float, optional): output array resolution.
                Defaults to 100.
            y_resolution (float, optional): output array resolution.
                Defaults to 100.
            x_bounds (tuple, optional): limits of output array.
                Defaults to None.
            y_bounds (tuple, optional): limits of output array.
                Defaults to None.

        Returns:
            :py:class:`floris.tools.flow_data.FlowData`: FlowData object at hub
                height.
        """
        if self.floris.farm.flow_field.wake.velocity_model.requires_resolution:
            raise (
                'Not allowed for wake model %s ' %
                self.floris.farm.flow_field.wake.velocity_model.model_string)

        # Get a copy for the flow field so don't change underlying grid points
        flow_field = copy.deepcopy(self.floris.farm.flow_field)

        # If x and y bounds are not provided, use rules of thumb
        if x_bounds is None:
            coords = self.floris.farm.flow_field.turbine_map.coords
            max_diameter = self.floris.farm.flow_field.max_diameter
            x = [coord.x1 for coord in coords]
            x_bounds = (min(x) - 2 * max_diameter, max(x) + 10 * max_diameter)
        if y_bounds is None:
            coords = self.floris.farm.flow_field.turbine_map.coords
            max_diameter = self.floris.farm.flow_field.max_diameter
            y = [coord.x2 for coord in coords]
            y_bounds = (min(y) - 2 * max_diameter, max(y) + 2 * max_diameter)

        # Z_bounds is always hub-height
        hub_height = self.floris.farm.flow_field.turbine_map.turbines[
            0].hub_height
        bounds_to_set = (x_bounds[0], x_bounds[1], y_bounds[0], y_bounds[1],
                         hub_height - 5., hub_height + 5.)

        # Set new bounds
        flow_field.set_bounds(bounds_to_set=bounds_to_set)

        # Change the resolution
        flow_field.reinitialize_flow_field(
            with_resolution=Vec3(x_resolution, y_resolution, 3))

        # Calculate the wakes
        flow_field.calculate_wake()

        order = "f"
        x = flow_field.x.flatten(order=order)
        y = flow_field.y.flatten(order=order)
        z = flow_field.z.flatten(order=order)

        u = flow_field.u.flatten(order=order)
        v = flow_field.v.flatten(order=order)
        w = flow_field.w.flatten(order=order)

        # Determine spacing, dimensions and origin
        unique_x = np.sort(np.unique(x))
        unique_y = np.sort(np.unique(y))
        unique_z = np.sort(np.unique(z))
        spacing = Vec3(unique_x[1] - unique_x[0], unique_y[1] - unique_y[0],
                       unique_z[1] - unique_z[0])
        dimensions = Vec3(len(unique_x), len(unique_y), len(unique_z))
        origin = Vec3(0.0, 0.0, 0.0)
        return FlowData(x,
                        y,
                        z,
                        u,
                        v,
                        w,
                        spacing=spacing,
                        dimensions=dimensions,
                        origin=origin)

    def get_flow_data(self, resolution=None, grid_spacing=10):
        """
        Generate FlowData object corresponding to the floris instance.

        #TODO disambiguate between resolution and grid spacing.

        Args:
            resolution (float, optional): resolution of output data.
                Only used for wake models that require spatial
                resolution (e.g. curl). Defaults to None.
            grid_spacing (int, optional): resolution of output data.
                Defaults to 10.

        Returns:
            :py:class:`floris.tools.flow_data.FlowData`: FlowData object
        """

        if resolution is None:
            if not self.floris.farm.flow_field.wake.velocity_model.requires_resolution:
                print('Assuming grid with spacing %d' % grid_spacing)
                xmin, xmax, ymin, ymax, zmin, zmax = self.floris.farm.flow_field.domain_bounds
                resolution = Vec3(1 + (xmax - xmin) / grid_spacing,
                                  1 + (ymax - ymin) / grid_spacing,
                                  1 + (zmax - zmin) / grid_spacing)
            else:
                print('Assuming model resolution')
                resolution = self.floris.farm.flow_field.wake.velocity_model.model_grid_resolution

        # Get a copy for the flow field so don't change underlying grid points
        flow_field = copy.deepcopy(self.floris.farm.flow_field)

        if flow_field.wake.velocity_model.requires_resolution and \
            flow_field.wake.velocity_model.model_grid_resolution != resolution:
            print(
                "WARNING: The current wake velocity model contains a required grid resolution;"
            )
            print(
                "    The Resolution given to FlorisInterface.get_flow_field is ignored."
            )
            resolution = flow_field.wake.velocity_model.model_grid_resolution
        flow_field.reinitialize_flow_field(with_resolution=resolution)
        print(resolution)
        flow_field.calculate_wake()

        order = "f"
        x = flow_field.x.flatten(order=order)
        y = flow_field.y.flatten(order=order)
        z = flow_field.z.flatten(order=order)

        u = flow_field.u.flatten(order=order)
        v = flow_field.v.flatten(order=order)
        w = flow_field.w.flatten(order=order)

        # Determine spacing, dimensions and origin
        unique_x = np.sort(np.unique(x))
        unique_y = np.sort(np.unique(y))
        unique_z = np.sort(np.unique(z))
        spacing = Vec3(unique_x[1] - unique_x[0], unique_y[1] - unique_y[0],
                       unique_z[1] - unique_z[0])
        dimensions = Vec3(len(unique_x), len(unique_y), len(unique_z))
        origin = Vec3(0.0, 0.0, 0.0)
        return FlowData(x,
                        y,
                        z,
                        u,
                        v,
                        w,
                        spacing=spacing,
                        dimensions=dimensions,
                        origin=origin)

    def get_yaw_angles(self):
        """
        Report yaw angles of wind turbines from instance of floris.

        Returns:
            yaw_angles (np.array): wind turbine yaw angles.
        """
        yaw_angles = [
            turbine.yaw_angle
            for turbine in self.floris.farm.turbine_map.turbines
        ]
        return yaw_angles

    def get_farm_power(self):
        """
        Report wind plant power from instance of floris.

        Returns:
            plant_power (float): sum of wind turbine powers.
        """
        turb_powers = [turbine.power for turbine in self.floris.farm.turbines]
        return np.sum(turb_powers)

    def get_turbine_power(self):
        """
        Report power from each wind turbine from instance of floris.

        Returns:
            turb_powers (np.array): power produced by each wind turbine.
        """
        turb_powers = [
            turbine.power
            for turbine in self.floris.farm.flow_field.turbine_map.turbines
        ]
        return turb_powers

    def get_turbine_ct(self):
        """
        Report thrust coefficient from each wind turbine from instance of floris.

        Returns:
            turb_ct_array (np.array): thrust coefficient for each wind turbine.
        """
        turb_ct_array = [
            turbine.Ct
            for turbine in self.floris.farm.flow_field.turbine_map.turbines
        ]
        return turb_ct_array

        # calculate the power under different yaw angles
    def get_power_for_yaw_angle_opt(self, yaw_angles):
        """
        Assign yaw angles to turbines, calculate wake, report power

        Args:
            yaw_angles (np.array): yaw to apply to each turbine

        Returns:
            power (float): wind plant power. #TODO negative? in kW?
        """

        self.calculate_wake(yaw_angles=yaw_angles)
        # self.floris.farm.set_yaw_angles(yaw_angles, calculate_wake=True)

        power = -1 * np.sum(
            [turbine.power for turbine in self.floris.farm.turbines])

        return power / (10**3)

    @property
    def layout_x(self):
        """
        Wind turbine coordinate information.

        Returns:
            layout_x (np.array): Wind turbine x-coordinate (east-west).
        """
        coords = self.floris.farm.flow_field.turbine_map.coords
        layout_x = np.zeros(len(coords))
        for i, coord in enumerate(coords):
            layout_x[i] = coord.x1prime
        return layout_x

    @property
    def layout_y(self):
        """
        Wind turbine coordinate information.

        Returns:
            layout_y (np.array): Wind turbine y-coordinate (east-west).
        """
        coords = self.floris.farm.flow_field.turbine_map.coords
        layout_y = np.zeros(len(coords))
        for i, coord in enumerate(coords):
            layout_y[i] = coord.x2prime
        return layout_y


# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np
from ..utilities import cosd, sind, tand


class WakeDeflection():
    """
    Base WakeDeflection object class. Subclasses are:

     - Jimenez
     - Gauss
     - Curl

    Each subclass has specific functional requirements. Refer to the
    each WakeDeflection subclass for further detail.
    """

    def __init__(self, parameter_dictionary):
        self.model_string = None

    def __str__(self):
        return self.model_string


class Jimenez(WakeDeflection):
    """
    Subclass of the
    :py:class:`floris.simulation.wake_deflection.WakeDeflection`
    object class. Parameters required for Jimenez wake model:

     - ad: #TODO What is this parameter for?
     - kd: #TODO What is this parameter for?
     - bd: #TODO What is this parameter for?
    """

    def __init__(self, parameter_dictionary):
        """
        Instantiate Jimenez object and pass function paramter values.

        Args:
            parameter_dictionary (dict): input dictionary with the
                following key-value pairs:
                    {
                        "kd": 0.05,
                        "ad": 0.0,
                        "bd": 0.0
                    }
        """
        super().__init__(parameter_dictionary)
        self.model_string = "jimenez"
        model_dictionary = parameter_dictionary[self.model_string]
        self.ad = float(model_dictionary["ad"])
        self.kd = float(model_dictionary["kd"])
        self.bd = float(model_dictionary["bd"])

    def function(self, x_locations, y_locations, turbine, coord, flow_field):
        """
        This function defines the angle at which the wake deflects in
        relation to the yaw of the turbine. This is coded as defined in
        the Jimenez et. al. paper.

        Args:
            x_locations (np.array): streamwise locations in wake
            y_locations (np.array): spanwise locations in wake
            turbine (:py:class:`floris.simulation.turbine.Turbine`):
                Turbine object
            coord
                (:py:meth:`floris.simulation.turbine_map.TurbineMap.coords`): 
                Spatial coordinates of wind turbine.
            flow_field
                (:py:class:`floris.simulation.flow_field.FlowField`): 
                Flow field object.

        Returns:
            deflection (np.array): Deflected wake centerline.
        """

        # angle of deflection
        xi_init = cosd(turbine.yaw_angle) * sind(
            turbine.yaw_angle) * turbine.Ct / 2.0

        x_locations = x_locations - coord.x1

        # yaw displacement
        yYaw_init = ( xi_init
            * ( 15 * (2 * self.kd * x_locations / turbine.rotor_diameter + 1)**4. + xi_init**2. )
            / ((30 * self.kd / turbine.rotor_diameter) * (2 * self.kd * x_locations / turbine.rotor_diameter + 1)**5.)) \
            - (xi_init * turbine.rotor_diameter * (15 + xi_init**2.) / (30 * self.kd))

        # corrected yaw displacement with lateral offset
        deflection = yYaw_init + self.ad + self.bd * x_locations

        x = np.unique(x_locations)
        for i in range(len(x)):
            tmp = np.max(deflection[x_locations == x[i]])
            deflection[x_locations == x[i]] = tmp

        return deflection


class Gauss(WakeDeflection):
    """
    Subclass of the
    :py:class:`floris.simulation.wake_deflection.WakeDeflection`
    object. Parameters required for Gauss wake model:

     - ka: #TODO What is this parameter for?
     - kb: #TODO What is this parameter for?
     - alpha: #TODO What is this parameter for?
     - beta: #TODO What is this parameter for?
     - ad: #TODO What is this parameter for?
     - bd: #TODO What is this parameter for?
    """

    def __init__(self, parameter_dictionary):
        """
        Instantiate Gauss object and pass function paramter values.

        Args:
            parameter_dictionary (dict): input dictionary with the
                following key-value pairs:
                    {
                        "ka": 0.3,
                        "kb": 0.004,
                        "alpha": 0.58,
                        "beta": 0.077,
                        "ad": 0.0,
                        "bd": 0.0
                    }
        """
        super().__init__(parameter_dictionary)
        self.model_string = "gauss"
        model_dictionary = parameter_dictionary[self.model_string]
        self.ka = float(model_dictionary["ka"])
        self.kb = float(model_dictionary["kb"])
        self.ad = float(model_dictionary["ad"])
        self.bd = float(model_dictionary["bd"])
        self.alpha = float(model_dictionary["alpha"])
        self.beta = float(model_dictionary["beta"])
        self.deflection_multiplier = 1.2

    def function(self, x_locations, y_locations, turbine, coord, flow_field):
        """
        This function defines the angle at which the wake deflects in
        relation to the yaw of the turbine. This is coded as defined in
        the Bastankah and Porte_Agel paper.

        Args:
            x_locations (np.array): streamwise locations in wake
            y_locations (np.array): spanwise locations in wake
            turbine (:py:class:`floris.simulation.turbine.Turbine`):
                Turbine object
            coord 
                (:py:meth:`floris.simulation.turbine_map.TurbineMap.coords`):
                Spatial coordinates of wind turbine.
            flow_field
                (:py:class:`floris.simulation.flow_field.FlowField`):
                Flow field object.

        Returns:
            deflection (np.array): Deflected wake centerline.
        """
        # ==============================================================
        wind_speed = flow_field.wind_speed  # free-stream velocity (m/s)
        TI_0 = flow_field.turbulence_intensity  # turbulence intensity (%/100)
        veer = flow_field.wind_veer  # veer (degrees)
        TI = TI_0

        # hard-coded model input data (goes in input file)
        ka = self.ka  # wake expansion parameter
        kb = self.kb  # wake expansion parameter
        alpha = self.alpha  # near wake parameter
        beta = self.beta  # near wake parameter
        ad = self.ad  # natural lateral deflection parameter
        bd = self.bd  # natural lateral deflection parameter

        # turbine parameters
        D = turbine.rotor_diameter
        yaw = -turbine.yaw_angle  # opposite sign convention in this model
        tilt = turbine.tilt_angle
        Ct = turbine.Ct

        # U_local = flow_field.wind_speed # just a placeholder for now, should be initialized with the flow_field
        U_local = flow_field.u_initial

        # initial velocity deficits
        uR = U_local * Ct * cosd(tilt) * cosd(yaw) / (
            2. * (1 - np.sqrt(1 - (Ct * cosd(tilt) * cosd(yaw)))))
        u0 = U_local * np.sqrt(1 - Ct)

        # length of near wake
        x0 = D * (cosd(yaw) * (1 + np.sqrt(1 - Ct * cosd(yaw)))) / (
            np.sqrt(2) * (4 * alpha * TI + 2 * beta *
                          (1 - np.sqrt(1 - Ct)))) + coord.x1

        # wake expansion parameters
        ky = ka * TI + kb
        kz = ka * TI + kb

        C0 = 1 - u0 / wind_speed
        M0 = C0 * (2 - C0)
        E0 = C0**2 - 3 * np.exp(1. / 12.) * C0 + 3 * np.exp(1. / 3.)

        # initial Gaussian wake expansion
        sigma_z0 = D * 0.5 * np.sqrt(uR / (U_local + u0))
        sigma_y0 = sigma_z0 * cosd(yaw) * cosd(veer)

        yR = y_locations - coord.x2
        xR = yR * tand(yaw) + coord.x1

        # yaw parameters (skew angle and distance from centerline)
        theta_c0 = self.deflection_multiplier * (
            0.3 * np.radians(yaw) / cosd(yaw)) * (
                1 - np.sqrt(1 - Ct * cosd(yaw)))  # skew angle in radians
        delta0 = np.tan(theta_c0) * (
            x0 - coord.x1
        )  # initial wake deflection; NOTE: use np.tan here since theta_c0 is radians

        # deflection in the near wake
        delta_near_wake = ((x_locations - xR) /
                           (x0 - xR)) * delta0 + (ad + bd *
                                                  (x_locations - coord.x1))
        delta_near_wake[x_locations < xR] = 0.0
        delta_near_wake[x_locations > x0] = 0.0

        # deflection in the far wake
        sigma_y = ky * (x_locations - x0) + sigma_y0
        sigma_z = kz * (x_locations - x0) + sigma_z0
        sigma_y[x_locations < x0] = sigma_y0[x_locations < x0]
        sigma_z[x_locations < x0] = sigma_z0[x_locations < x0]

        ln_deltaNum = (1.6 + np.sqrt(M0)) * (
            1.6 * np.sqrt(sigma_y * sigma_z /
                          (sigma_y0 * sigma_z0)) - np.sqrt(M0))
        ln_deltaDen = (1.6 - np.sqrt(M0)) * (
            1.6 * np.sqrt(sigma_y * sigma_z /
                          (sigma_y0 * sigma_z0)) + np.sqrt(M0))
        delta_far_wake = delta0 + (theta_c0 * E0 / 5.2) * np.sqrt(
            sigma_y0 * sigma_z0 /
            (ky * kz * M0)) * np.log(ln_deltaNum / ln_deltaDen) + (
                ad + bd * (x_locations - coord.x1))
        delta_far_wake[x_locations <= x0] = 0.0

        deflection = delta_near_wake + delta_far_wake

        return deflection


class Curl(WakeDeflection):
    """
    Subclass of the
    :py:class:`floris.simulation.wake_deflection.WakeDeflection`
    object. Parameters required for Curl wake model:

     - model_grid_resolution: #TODO What does this do?
    """

    def __init__(self, parameter_dictionary):
        """
        Instantiate Curl object and pass function paramter values.

        Args:
            parameter_dictionary (dict): input dictionary with the
                following key-value pair:
                    {
                        "model_grid_resolution": [
                                                    250,
                                                    100,
                                                    75
                                                ],
                    }
        """
        super().__init__(parameter_dictionary)
        self.model_string = "curl"

    def function(self, x_locations, y_locations, turbine, coord, flow_field):
        """
        This function will return the wake centerline predicted with
        the curled wake model. #TODO Eventually. This is coded as
        defined in the Martinez-Tossas et al. paper.
        """
        return np.zeros(np.shape(x_locations))



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from ..utilities import Vec3
import matplotlib.pyplot as plt
from matplotlib import rcParams
import numpy as np


def plot_turbines(ax, layout_x, layout_y, yaw_angles, D):
    """
    Plot wind plant layout from turbine locations.

    Args:
        ax (:py:class:`matplotlib.pyplot.axes`): figure axes.
        layout_x (np.array): wind turbine locations (east-west).
        layout_y (np.array): wind turbine locations (north-south).
        yaw_angles (np.array): yaw angles of each wind turbine.
        D (float): wind turbine rotor diameter.
    """
    for x, y, yaw in zip(layout_x, layout_y, yaw_angles):
        R = D / 2.
        x_0 = x + np.sin(np.deg2rad(yaw)) * R
        x_1 = x - np.sin(np.deg2rad(yaw)) * R
        y_0 = y - np.cos(np.deg2rad(yaw)) * R
        y_1 = y + np.cos(np.deg2rad(yaw)) * R
        ax.plot([x_0, x_1], [y_0, y_1], color='k')


def line_contour_cut_plane(cut_plane,
                           ax=None,
                           levels=None,
                           colors=None,
                           **kwargs):
    """
    Visualize the scan as a simple contour.

    Args:
        cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`): 
            CutPlane Object.
        ax (:py:class:`matplotlib.pyplot.axes`): figure axes. Defaults 
            to None.
        levels (np.array, optional): contour levels for plot.
            Defaults to None.
        colors (list, optional): strings of color specification info.
            Defaults to None.
    """

    if not ax:
        fig, ax = plt.subplots()

    # Reshape UMesh internally
    u_mesh = cut_plane.u_mesh.reshape(cut_plane.resolution[1],
                                      cut_plane.resolution[0])
    Zm = np.ma.masked_where(np.isnan(u_mesh), u_mesh)
    rcParams['contour.negative_linestyle'] = 'solid'

    # # Plot the cut-through
    ax.contour(cut_plane.x1_lin,
               cut_plane.x2_lin,
               Zm,
               levels=levels,
               colors=colors,
               **kwargs)

    # Make equal axis
    ax.set_aspect('equal')


def visualize_cut_plane(cut_plane,
                        ax=None,
                        minSpeed=None,
                        maxSpeed=None,
                        cmap='coolwarm'):
    """
    Generate pseudocolor mesh plot of the scan.

    Args:
        cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`): 2D 
            plane through wind plant.
        ax (:py:class:`matplotlib.pyplot.axes`): figure axes. Defaults 
            to None.
        minSpeed (float, optional): Minimum value of wind speed for
            contours. Defaults to None.
        maxSpeed (float, optional): Maximum value of wind speed for
            contours. Defaults to None.
        cmap (str, optional): Colormap specifier. Defaults to 
            'coolwarm'.

    Returns:
        im (plt.pcolormesh): image handle
    """

    if not ax:
        fig, ax = plt.subplots()
    if minSpeed is None:
        minSpeed = cut_plane.u_mesh.min()
    if maxSpeed is None:
        maxSpeed = cut_plane.u_mesh.max()

    # Reshape UMesh internally
    u_mesh = cut_plane.u_mesh.reshape(cut_plane.resolution[1],
                                      cut_plane.resolution[0])
    Zm = np.ma.masked_where(np.isnan(u_mesh), u_mesh)

    # Plot the cut-through
    im = ax.pcolormesh(cut_plane.x1_lin,
                       cut_plane.x2_lin,
                       Zm,
                       cmap=cmap,
                       vmin=minSpeed,
                       vmax=maxSpeed)

    # Add line contour
    line_contour_cut_plane(cut_plane,
                           ax=ax,
                           levels=None,
                           colors='w',
                           linewidths=0.8,
                           alpha=0.3)

    # Make equal axis
    ax.set_aspect('equal')

    # Return im
    return im


def visualize_quiver(cut_plane,ax=None,minSpeed=None,maxSpeed=None,downSamp=1,**kw):
        """ Visualize the scan
        
        Args:
            ax: axes for plotting, if none, create a new one  
            minSpeed, maxSpeed, values used for plotting, if not provide assume to data max min
        """

        """
        Visualize the in-plane flows in a cut_plane

        Args:
            cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`): 2D 
                plane through wind plant.
            ax (:py:class:`matplotlib.pyplot.axes`): figure axes. Defaults 
                to None.
            minSpeed (float, optional): Minimum value of wind speed for
                contours. Defaults to None.
            maxSpeed (float, optional): Maximum value of wind speed for
                contours. Defaults to None.
            downSamp (int, optional): downSamp the number of quiver arrows from underlying grid

        Returns:
            im (plt.pcolormesh): image handle
        """
        if not ax:
            fig, ax = plt.subplots()


        # # Reshape UMesh internally
        v_mesh = cut_plane.v_mesh.reshape(cut_plane.resolution[1],
                                      cut_plane.resolution[0])
        w_mesh = cut_plane.w_mesh.reshape(cut_plane.resolution[1],
                                      cut_plane.resolution[0])
        # Zm = np.ma.masked_where(np.isnan(uMesh),uMesh)

        # plot the stream plot
        QV1 = ax.quiver( (cut_plane.x1_mesh[::downSamp,::downSamp]),
                   (cut_plane.x2_mesh[::downSamp,::downSamp]),
                   v_mesh[::downSamp,::downSamp],
                   w_mesh[::downSamp,::downSamp],
                   scale=80.0,alpha=0.75,
                   **kw)

        ax.quiverkey(QV1, -.75, -0.4, 1, '1 m/s', coordinates='data')
        # ax.quiverkey(QV1, -3, 1.2, 1, '1 m/s', coordinates='data')

        #print(minSpeed,maxSpeed)
        
        # Make equal axis
        ax.set_aspect('equal')

def reverse_cut_plane_x_axis_in_plot(ax):
    """
    Shortcut method to reverse direction of x-axis.

    Args:
        ax (:py:class:`matplotlib.pyplot.axes`): figure axes.
    """
    ax.invert_xaxis()


# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np
from scipy.interpolate import interp1d
from scipy.interpolate import griddata
from ..utilities import cosd, sind, tand


class Turbine():
    """
    Turbine is a class containing objects pertaining to the individual 
    turbines.

    Turbine is a model class representing a particular wind turbine. It 
    is largely a container of data and parameters, but also contains 
    methods to probe properties for output.

    Args:
        instance_dictionary: A dictionary that is generated from the 
            input_reader; it should have the following key-value pairs:

            -   **description**: A string containing a description of 
                the turbine.
            -   **properties**: A dictionary containing the following 
                key-value pairs:

                -   **rotor_diameter**: A float that is the rotor 
                    diameter (m/s).
                -   **hub_height**: A float that is the hub height (m).
                -   **blade_count**: An integer that is the number of 
                    blades.
                -   **pP**: A float that is the cosine exponent 
                    relating the yaw misalignment angle to power.
                -   **pT**: A float that is the cosine exponent 
                    relating the rotor tilt angle to power.
                -   **generator_efficiency**: A float that is the 
                    generator efficiency factor used to scale the power 
                    production.
                -   **power_thrust_table**: A dictionary containing the 
                    following key-value pairs:

                    -   **power**: A list of floats describing the 
                        coefficient of power at different wind speeds.
                    -   **thrust**: A list of floats describing the 
                        coefficient of thrust at different wind speeds.
                    -   **wind_speed**: A list of floats containing the 
                        wind speeds for which the power and thrust 
                        values are provided (m/s).

                -   **yaw_angle**: A float that is the yaw angle of the 
                    turbine relative to the wind direction (deg). A 
                    positive value represents a counter-clockwise 
                    rotation relative to the wind direction. 
                -   **tilt_angle**: A float that is the tilt angle of 
                    the turbine (deg). Positive values correspond to a 
                    downward rotation of the rotor for an upstream 
                    turbine. 
                -   **TSR**: A float that is the tip-speed ratio of the 
                    turbine. This parameter is used in the "curl" wake 
                    model.

    Returns:
        Turbine: An instantiated Turbine object.
    """

    def __init__(self, instance_dictionary):

        self.description = instance_dictionary["description"]
        properties = instance_dictionary["properties"]
        self.rotor_diameter = properties["rotor_diameter"]
        self.hub_height = properties["hub_height"]
        self.blade_count = properties["blade_count"]
        self.pP = properties["pP"]
        self.pT = properties["pT"]
        self.generator_efficiency = properties["generator_efficiency"]
        self.power_thrust_table = properties["power_thrust_table"]
        self.yaw_angle = properties["yaw_angle"]
        self.tilt_angle = properties["tilt_angle"]
        self.tsr = properties["TSR"]

        # constants
        self.grid_point_count = 5*5
        if np.sqrt(self.grid_point_count) % 1 != 0.0:
            raise ValueError(
                "Turbine.grid_point_count must be the square of a number")

        self.reinitialize_turbine(turbulence_intensity=None)

        # initialize derived attributes
        self.grid = self._create_swept_area_grid()

        # initialize to an invalid value until calculated
        self.air_density = -1

    # Private methods

    def _create_swept_area_grid(self):
        # TODO: add validity check:
        # rotor points has a minimum in order to always include points inside
        # the disk ... 2?
        #
        # the grid consists of the y,z coordinates of the discrete points which
        # lie within the rotor area: [(y1,z1), (y2,z2), ... , (yN, zN)]

        # update:
        # using all the grid point because that how roald did it.
        # are the points outside of the rotor disk used later?

        # determine the dimensions of the square grid
        num_points = int(np.round(np.sqrt(self.grid_point_count)))
        # syntax: np.linspace(min, max, n points)
        horizontal = np.linspace(-self.rotor_radius,
                                 self.rotor_radius, num_points)
        vertical = np.linspace(-self.rotor_radius,
                               self.rotor_radius, num_points)

        # build the grid with all of the points
        grid = [(h, vertical[i]) for i in range(num_points)
                for h in horizontal]

        # keep only the points in the swept area
        grid = [point for point in grid if np.hypot(
            point[0], point[1]) < self.rotor_radius]

        return grid

    def _fCp(self, at_wind_speed):
        cp = self.power_thrust_table["power"]
        wind_speed = self.power_thrust_table["wind_speed"]
        fCpInterp = interp1d(wind_speed, cp, fill_value='extrapolate')
        if at_wind_speed < min(wind_speed):
            return 0.
        else:
            _cp = fCpInterp(at_wind_speed)
            if _cp.size > 1:
                _cp = _cp[0]
            return float(_cp)

    def _fCt(self, at_wind_speed):
        ct = self.power_thrust_table["thrust"]
        wind_speed = self.power_thrust_table["wind_speed"]
        fCtInterp = interp1d(wind_speed, ct, fill_value='extrapolate')
        if at_wind_speed < min(wind_speed):
            return 0.99
        else:
            _ct = fCtInterp(at_wind_speed)
            if _ct.size > 1:
                _ct = _ct[0]
            if _ct > 1.0:
                _ct = 0.99
            return float(_ct)

    # Public methods

    def calculate_swept_area_velocities(self, wind_direction, local_wind_speed, coord, x, y, z):
        """
        This method calculates and returns the wind speeds at each 
        rotor swept area grid point for the turbine, interpolated from 
        the flow field grid.

        Args:
            wind_direction: A float that is the wind farm wind 
                direction (deg).
            local_wind_speed: An array of floats that contains the wind 
                speed at each grid point in the flow field (m/s).
            coord: A :py:obj:`floris.utilities.Vec3` object containing 
                the coordinate of the turbine.
            x: An array of floats containing the "x" coordinates of the 
                flow field grid.
            y: An array of floats containing the "y" coordinates of the 
                flow field grid.
            z: An array of floats containing the "z" coordinates of the 
                flow field grid.

        Returns:
            numpy.ndarray: A numpy array of floats containing the wind 
            speed at each rotor grid point for the turbine (m/s).
        """
        u_at_turbine = local_wind_speed
        x_grid = x
        y_grid = y
        z_grid = z

        yPts = np.array([point[0] for point in self.grid])
        zPts = np.array([point[1] for point in self.grid])

        # interpolate from the flow field to get the flow field at the grid points
        dist = [np.sqrt((coord.x1 - x_grid)**2 + (coord.x2 + yPts[i] - y_grid) **
                        2 + (self.hub_height + zPts[i] - z_grid)**2) for i in range(len(yPts))]
        idx = [np.where(dist[i] == np.min(dist[i])) for i in range(len(yPts))]
        data = [np.mean(u_at_turbine[idx[i]]) for i in range(len(yPts))]
        return np.array(data)

    def calculate_turbulence_intensity(self, flow_field_ti, velocity_model, turbine_coord, wake_coord, turbine_wake):
        """
        Calculates the turbulence intensity at a specific wind turbine.

        This method calculates and returns the turbulence intensity at 
        the wind turbine consisting of the ambient turbulence as well 
        as the wake-added turbulence from an upstream turbine, using 
        the approach of Crespo, A. and Herna, J. "Turbulence 
        characteristics in wind-turbine wakes." *J. Wind Eng Ind 
        Aerodyn*. 1996.

        Args:
            flow_field_ti: A float that is the ambient turbulence 
                intensity in the flow field expressed as a decimal 
                fraction.
            velocity_model: A 
                :py:obj:`floris.simulation.wake_velocity.WakeVelocity` 
                object containing wake model parameters.
            turbine_coord: A :py:obj:`floris.utilities.Vec3` object 
                containing the coordinate of the turbine.
            wake_coord: A :py:obj:`floris.utilities.Vec3` object 
                containing the coordinate of the upstream turbine.
            turbine_wake: A :py:class:`floris.simulation.turbine` 
                object that represents the upstream turbine.

        Returns:
            numpy.float64: The turbulence intensity at the current 
            turbine including ambient turbulence and turbulence added 
            by the upstream turbine wake.
        """

        ti_initial = flow_field_ti

        # user-input turbulence intensity parameters
        ti_i = velocity_model.ti_initial
        ti_constant = velocity_model.ti_constant
        ti_ai = velocity_model.ti_ai
        ti_downstream = velocity_model.ti_downstream

        # turbulence intensity calculation based on Crespo et. al.
        ti_calculation = ti_constant \
            * turbine_wake.aI**ti_ai \
            * ti_initial**ti_i \
            * ((turbine_coord.x1 - wake_coord.x1) / self.rotor_diameter)**ti_downstream

        return np.sqrt(ti_calculation**2 + flow_field_ti**2)

    def update_velocities(self, u_wake, coord, flow_field, rotated_x, rotated_y, rotated_z):
        """
        This method updates the velocities at the rotor swept area grid 
        points based on the flow field freestream velocities and wake 
        velocities.

        Args:
            u_wake: An array of floats containing the wake deficit 
                velocities at all grid points in the flow field (m/s).
            coord: A :py:obj:`floris.utilities.Vec3` object containing 
                the coordinate of the turbine.
            flow_field: A :py:class:`floris.simulation.flow_field` 
                object.
            rotated_x: An array of floats containing the "x" 
                coordinates of the flow field grid rotated so the new 
                "x" axis is aligned with the wind direction.
            rotated_y: An array of floats containing the "y" 
                coordinates of the flow field grid rotated so the new 
                "x" axis is aligned with the wind direction.
            rotated_z: An array of floats containing the "z" 
                coordinates of the flow field grid rotated so the new 
                "x" axis is aligned with the wind direction.

        Returns:
            *None* -- The velocities are updated directly in the 
            :py:class:`floris.simulation.turbine` object.
        """

        # reset the waked velocities
        local_wind_speed = flow_field.u_initial - u_wake
        self.velocities = self.calculate_swept_area_velocities(
            flow_field.wind_direction,
            local_wind_speed,
            coord,
            rotated_x,
            rotated_y,
            rotated_z
        )

    def reinitialize_turbine(self, turbulence_intensity):
        """
        This method sets the velocities at the turbine's rotor swept 
        area grid points to zero.

        Returns:
            *None* -- The velocities are updated directly in the 
            :py:class:`floris.simulation.turbine` object.
        """
        self.velocities = [0.0] * self.grid_point_count
        self.turbulence_intensity = turbulence_intensity

    def set_yaw_angle(self, yaw_angle):
        """
        This method sets the turbine's yaw angle.

        Args:
            yaw_angle: A float that is the new yaw angle (deg).

        Returns:
            **None** -- The yaw angle is stored in the 
            :py:class:`floris.simulation.turbine` object.

        Examples:
            To set a turbine's yaw angle:

            >>> floris.farm.turbines[0].set_yaw_angle(20.0)
        """
        self._yaw_angle = yaw_angle

    # Getters & Setters

    @property
    def rotor_radius(self):
        """
        This method returns the rotor radius of the turbine (m).

        Returns:
            float: The rotor radius of the turbine.

        Examples:
            To get the rotor radius for a turbine:

            >>> rotor_radius = floris.farm.turbines[0].rotor_radius()
        """
        return self.rotor_diameter / 2.0

    @property
    def yaw_angle(self):
        """
        This method gets or sets the turbine's yaw angle.

        Args:
            value: A float that is the new yaw angle (deg).

        Returns:
            float: The current yaw angle (deg).

        Examples:
            To set the yaw angle for each turbine in the wind farm:

            >>> yaw_angles = [20.0, 10.0, 0.0]
            >>> for yaw_angle, turbine in 
            ... zip(yaw_angles, floris.farm.turbines):
            ...     turbine.yaw_angle = yaw_angle

            To get the current yaw angle for each turbine in the wind 
            farm:

            >>> yaw_angles = []
            >>> for i, turbine in enumerate(floris.farm.turbines):
            ...     yaw_angles.append(turbine.yaw_angle())
        """
        return self._yaw_angle

    @yaw_angle.setter
    def yaw_angle(self, value):
        self._yaw_angle = value

    @property
    def tilt_angle(self):
        """
        This method gets the turbine's tilt angle.

        Args:
            value: A float that is the new tilt angle (deg).

        Returns:
            float: The current tilt angle (deg).

        Examples:
            To get the current tilt angle for a turbine:

            >>> tilt_angle = floris.farm.turbines[0].tilt_angle()
        """
        return self._tilt_angle

    @tilt_angle.setter
    def tilt_angle(self, value):
        self._tilt_angle = value

    @property
    def average_velocity(self):
        """
        This property calculates and returns the cube root of the 
        mean cubed velocity in the turbine's rotor swept area (m/s).

        Returns:
            numpy.float64: The average velocity across a rotor.

        Examples:
            To get the average velocity for a turbine:

            >>> avg_vel = floris.farm.turbines[0].average_velocity()
        """
        return np.cbrt(np.mean(self.velocities**3))


    @property
    def Cp(self):
        """
        This property returns the power coeffcient of a turbine.

        This property returns the coefficient of power of the turbine 
        using the rotor swept area average velocity, interpolated from 
        the coefficient of power table. The average velocity is 
        calculated as the cube root of the mean cubed velocity in the 
        rotor area.

        Note, the velocity is scalled to an effective velocity by the yaw

        Returns:
            float: The power coefficient of a turbine at the current 
            operating conditions.

        Examples:
            To get the power coefficient value for a turbine:

            >>> Cp = floris.farm.turbines[0].Cp()
        """
        # Compute the yaw effective velocity
        pW = self.pP / 3.0 # Convert from pP to pW
        yaw_effective_velocity = self.average_velocity * cosd(self.yaw_angle) ** pW

        return self._fCp(yaw_effective_velocity)

    @property
    def Ct(self):
        """
        This property returns the thrust coefficient of a turbine.

        This method returns the coefficient of thrust of the yawed 
        turbine, interpolated from the coefficient of power table, 
        using the rotor swept area average velocity and the turbine's 
        yaw angle. The average velocity is calculated as the cube root 
        of the mean cubed velocity in the rotor area.

        Returns:
            float: The thrust coefficient of a turbine at the current 
            operating conditions.

        Examples:
            To get the thrust coefficient value for a turbine:

            >>> Ct = floris.farm.turbines[0].Ct()
        """
        return self._fCt(self.average_velocity) * cosd(self.yaw_angle)# **self.pP

    @property
    def power(self):
        """
        This property returns the power produced by turbine (W), 
        adjusted for yaw and tilt.

        Returns:
            float: Power of a turbine in watts.

        Examples:
            To get the power for a turbine:

            >>> power = floris.farm.turbines[0].power()
        """

        # Update to power calculation which replaces the fixed pP exponent with
        # an exponent pW, that changes the effective wind speed input to the power 
        # calculation, rather than scaling the power.  This better handles power
        # loss to yaw in above rated conditions
        # 
        # based on the paper "Optimising yaw control at wind farm level" by
        # Ervin Bossanyi

        # Compute the yaw effective velocity
        pW = self.pP / 3.0 # Convert from pP to w
        yaw_effective_velocity = self.average_velocity * cosd(self.yaw_angle) ** pW
        
        # Now compute the power
        cptmp = self.Cp #Note Cp is also now based on yaw effective velocity
        return 0.5 * self.air_density * (np.pi * self.rotor_radius**2) \
            * cptmp * self.generator_efficiency \
            * yaw_effective_velocity**3

    @property
    def aI(self):
        """
        This property returns the axial induction factor of the yawed 
        turbine calculated from the coefficient of thrust and the yaw 
        angle.

        Returns:
            float: Axial induction factor of a turbine.

        Examples:
            To get the axial induction factor for a turbine:

            >>> aI = floris.farm.turbines[0].aI()
        """
        return 0.5 / cosd(self.yaw_angle) \
            * (1 - np.sqrt(1 - self.Ct * cosd(self.yaw_angle)))



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import os
import pickle
from floris.utilities import wrap_180


class PowerRose():
    """
    PowerRose object class used to parse data and generate figures.
    """

    def __init__(self):
        """
        Init Function of WindRose Object.
        No explicit arguments required.
        """

    def _all_combine(self):
        df_power = self.df_power.copy(deep=True)
        df_yaw = self.df_yaw.copy(deep=True)
        df_turbine_power_no_wake = self.df_turbine_power_no_wake.copy(
            deep=True)
        df_turbine_power_baseline = self.df_turbine_power_baseline.copy(
            deep=True)
        df_turbine_power_opt = self.df_turbine_power_opt.copy(deep=True)

        # Adjust the column names for uniqunes
        df_yaw.columns = [
            'yaw_%d' % c if type(c) is int else c for c in df_yaw.columns
        ]
        df_turbine_power_no_wake.columns = [
            'tnw_%d' % c if type(c) is int else c
            for c in df_turbine_power_no_wake.columns
        ]
        df_turbine_power_baseline.columns = [
            'tb_%d' % c if type(c) is int else c
            for c in df_turbine_power_baseline.columns
        ]
        df_turbine_power_opt.columns = [
            'topt_%d' % c if type(c) is int else c
            for c in df_turbine_power_opt.columns
        ]

        # Merge
        df_combine = df_power.merge(df_yaw, on=['ws', 'wd'])
        df_combine = df_combine.merge(df_turbine_power_no_wake,
                                      on=['ws', 'wd'])
        df_combine = df_combine.merge(df_turbine_power_baseline,
                                      on=['ws', 'wd'])
        df_combine = df_combine.merge(df_turbine_power_opt, on=['ws', 'wd'])

        return df_combine

    def _norm_frequency(self, df):
        print('Norming frequency total of %.2f to 1.0' % df.freq_val.sum())
        df['freq_val'] = df.freq_val / df.freq_val.sum()
        return df

    def _compute_energy(self):
        self.df_power[
            'energy_no_wake'] = self.df_power.freq_val * self.df_power.power_no_wake
        self.df_power[
            'energy_baseline'] = self.df_power.freq_val * self.df_power.power_baseline
        self.df_power[
            'energy_opt'] = self.df_power.freq_val * self.df_power.power_opt

    def _compute_totals(self):
        df = self.df_power.copy(deep=True)
        df = df.sum()

        # Get total energy amounts
        self.total_no_wake = df.energy_no_wake
        self.total_baseline = df.energy_baseline
        self.total_opt = df.energy_opt

        # Get wake loss amounts
        self.baseline_percent = self.total_baseline / self.total_no_wake
        self.baseline_wake_loss = 1 - self.baseline_percent
        self.opt_percent = self.total_opt / self.total_no_wake
        self.opt_wake_loss = 1 - self.opt_percent

        # Percent gain
        self.percent_gain = (self.total_opt -
                             self.total_baseline) / self.total_baseline
        self.reduction_in_wake_loss = -1 * (
            self.opt_wake_loss -
            self.baseline_wake_loss) / self.baseline_wake_loss

    def report(self):
        """
        Print operational data information to screen.
        """

        print('=============================================')
        print('Case %s has results:' % self.name)
        print('=============================================')
        print('-\tNo-Wake\t\tBaseline\tOpt ')
        print('---------------------------------------------')
        print('Energy\t%.1E\t\t%.1E\t\t%.1E' %
              (self.total_no_wake, self.total_baseline, self.total_opt))
        print('%%\t--\t\t%.1f%%\t\t%.1f%%' %
              (100. * self.baseline_percent, 100. * self.opt_percent))
        print('Wk Loss\t--\t\t%.1f%%\t\t%.1f%%' %
              (100. * self.baseline_wake_loss, 100. * self.opt_wake_loss))
        print('AEP Gain --\t\t--\t\t%.1f%%' % (100. * self.percent_gain))
        print('Loss Red --\t\t--\t\t%.1f%%' %
              (100. * self.reduction_in_wake_loss))

    def initialize(self, name, df_power, df_yaw, df_turbine_power_no_wake,
                   df_turbine_power_baseline, df_turbine_power_opt):
        """
        Additional initialization of the PowerRose object.
        #TODO Shouldn't this just be part of the __init__ method?

        Args:
            name (str): name of PowerRose object.
            df_power (pd.DataFrame): plant power data.
            df_yaw (pd.DataFrame): yaw data.
            df_turbine_power_no_wake (pd.DataFrame): wind turbine power
                data without wake losses.
            df_turbine_power_baseline (pd.DataFrame): wind turbine
                baseline power data.
            df_turbine_power_opt (pd.DataFrame): optimal wind turbine
                power data.
        """
        self.name = name
        self.df_power = self._norm_frequency(df_power)
        self.df_yaw = df_yaw
        self.df_turbine_power_no_wake = df_turbine_power_no_wake
        self.df_turbine_power_baseline = df_turbine_power_baseline
        self.df_turbine_power_opt = df_turbine_power_opt

        # Make a single combined frame in case it's useful
        self.df_combine = self._all_combine()

        # Compute energies
        self._compute_energy()

        # Compute totals
        self._compute_totals()

    def load(self, filename):
        """
        Load PowerRose object from pickle file.

        Args:
            filename (str): Read-from path to pickled PowerRose object.
        """

        self.name, self.df_power, self.df_yaw, self.df_turbine_power_no_wake, self.df_turbine_power_baseline, self.df_turbine_power_opt, self.df_combine = pickle.load(
            open(filename, "rb"))

        # Compute energies
        self._compute_energy()

        # Compute totals
        self._compute_totals()

    def save(self, filename):
        """
        Pickle a PowerRose Object

        Args:
            filename (str): Write-to path for PowerRose pickle.
        """
        pickle.dump([
            self.name, self.df_power, self.df_yaw,
            self.df_turbine_power_no_wake, self.df_turbine_power_baseline,
            self.df_turbine_power_opt, self.df_combine
        ], open(filename, "wb"))

    def plot_by_direction(self):
        """
        Bin Power data by direction and plot.

        Returns:
            tuple: tuple containing:

                -   **fig** (*plt.figure*): Figure handle.
                -   **axarr** (*list*): list of axis handles.
        """

        df = self.df_power.copy(deep=True)
        df = df.groupby('wd').sum().reset_index()

        fig, axarr = plt.subplots(2, 1, sharex=True)
        ax = axarr[0]
        ax.plot(df.wd,
                df.energy_baseline / df.energy_no_wake,
                label='Baseline',
                color='k')
        ax.axhline(np.mean(df.energy_baseline) / np.mean(df.energy_no_wake),
                   color='k',
                   ls='--')
        ax.plot(df.wd,
                df.energy_opt / df.energy_no_wake,
                label='Optimized',
                color='r')
        ax.axhline(np.mean(df.energy_opt) / np.mean(df.energy_no_wake),
                   color='r',
                   ls='--')
        ax.set_ylabel('Wake Loss')
        ax.grid(True)
        ax.legend()

        ax = axarr[1]
        ax.plot(
            df.wd,
            100. * (df.energy_opt - df.energy_baseline) / df.energy_baseline,
            'r')
        ax.axhline(100. * (df.energy_opt.mean() - df.energy_baseline.mean()) /
                   df.energy_baseline.mean(),
                   color='r',
                   ls='--')
        ax.set_ylabel('Percent Gain')

        return fig, axarr

    def wake_loss_at_direction(self, wd):
        """
        Calculate wake losses for a given direction. Plot rose figures
        for Power, Energy, Baseline power, Optimal gain, Total Gain,
        Percent Gain, etc.

        Args:
            wd (float): Wind direction of interest.

        Returns:
            tuple: tuple containing:

                -   **fig** (*plt.figure*): Figure handle.
                -   **axarr** (*list*): list of axis handles.
        """

        df = self.df_power.copy(deep=True)

        # Choose the nearest direction
        # Find nearest wind direction
        df['dist'] = np.abs(wrap_180(df.wd - wd))
        wd_select = df[df.dist == df.dist.min()]['wd'].unique()[0]
        print('Nearest wd to %.1f is %.1f' % (wd, wd_select))
        df = df[df.wd == wd_select]

        df = df.groupby('ws').sum().reset_index()

        fig, axarr = plt.subplots(4, 2, sharex=True, figsize=(14, 12))

        ax = axarr[0, 0]
        ax.set_title('Power')
        ax.plot(df.ws, df.power_no_wake, 'k', label='No Wake')
        ax.plot(df.ws, df.power_baseline, 'b', label='Baseline')
        ax.plot(df.ws, df.power_opt, 'r', label='Opt')
        ax.set_ylabel('Total')
        ax.grid()

        ax = axarr[0, 1]
        ax.set_title('Energy')
        ax.plot(df.ws, df.energy_no_wake, 'k', label='No Wake')
        ax.plot(df.ws, df.energy_baseline, 'b', label='Baseline')
        ax.plot(df.ws, df.energy_opt, 'r', label='Opt')
        ax.legend()
        ax.grid()

        ax = axarr[1, 0]
        ax.plot(df.ws,
                df.power_baseline / df.power_no_wake,
                'b',
                label='Baseline')
        ax.plot(df.ws, df.power_opt / df.power_no_wake, 'r', label='Opt')
        ax.set_ylabel('Percent')
        ax.grid()

        ax = axarr[1, 1]
        ax.plot(df.ws,
                df.energy_baseline / df.energy_no_wake,
                'b',
                label='Baseline')
        ax.plot(df.ws, df.energy_opt / df.energy_no_wake, 'r', label='Opt')
        ax.grid()

        ax = axarr[2, 0]
        ax.plot(df.ws, (df.power_opt - df.power_baseline), 'r')
        ax.set_ylabel('Total Gain')
        ax.grid()

        ax = axarr[2, 1]
        ax.plot(df.ws, (df.energy_opt - df.energy_baseline), 'r')
        ax.grid()

        ax = axarr[3, 0]
        ax.plot(df.ws, (df.power_opt - df.power_baseline) / df.power_baseline,
                'r')
        ax.set_ylabel('Percent Gain')
        ax.grid()
        ax.set_xlabel('Wind Speed (m/s)')

        ax = axarr[3, 1]
        ax.plot(df.ws,
                (df.energy_opt - df.energy_baseline) / df.energy_baseline, 'r')
        ax.grid()
        ax.set_xlabel('Wind Speed (m/s)')

        return fig, axarr


# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np
from scipy.optimize import minimize

# import warnings
# warnings.simplefilter('ignore', RuntimeWarning)


def optimize_yaw(fi, minimum_yaw_angle=0.0, maximum_yaw_angle=25.0):
    """
    Find optimum setting of turbine yaw angles for power production
    given fixed atmospheric conditins (wind speed, direction, etc.)

    Args:
        fi (:py:class:`floris.tools.floris_utilities.FlorisInterface`):
            Interface from FLORIS to the wfc tools
        minimum_yaw_angle (float, optional): minimum constraint on yaw.
            Defaults to 0.0.
        maximum_yaw_angle (float, optional): maximum constraint on yaw.
            Defaults to 25.0.

    Returns:
        opt_yaw_angles (np.array): optimal yaw angles of each turbine.
    """
    # initialize floris without the full flow domain; only points assigned at the turbine
    fi.floris.farm.flow_field.reinitialize_flow_field()

    # set initial conditions
    x0 = []
    bnds = []

    turbines = fi.floris.farm.turbine_map.turbines
    x0 = [turbine.yaw_angle for turbine in turbines]
    bnds = [(minimum_yaw_angle, maximum_yaw_angle) for turbine in turbines]

    print('=====================================================')
    print('Optimizing wake redirection control...')
    print('Number of parameters to optimize = ', len(x0))
    print('=====================================================')

    residual_plant = minimize(fi.get_power_for_yaw_angle_opt,
                              x0,
                              method='SLSQP',
                              bounds=bnds,
                              options={'eps': np.radians(5.0)})

    if np.sum(residual_plant.x) == 0:
        print('No change in controls suggested for this inflow condition...')

    # %%
    opt_yaw_angles = residual_plant.x

    return opt_yaw_angles



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import griddata

def nudge_outward(x):
    """
    Avoid numerical issue in grid data by sligly expanding input x,y

    Args:
        x (np.arraym float): Vector to be slightly expanded
    """
    nudge_val = 0.001
    min_x = np.min(x)
    max_x = np.max(x)
    x = np.where(x==min_x, min_x-nudge_val, x) 
    x = np.where(x==max_x, max_x+nudge_val, x) 
    return x

class _CutPlane():
    def __init__(self, flow_data, x1='x', x2='y', x3_value=None):
        """
        Initialize CutPlane object. Used to extract a 2D plane from a
        3D vectoral velocity field

        Args:
            flow_data (np.array): 3D vector field of velocity data
            x1 (str, optional): first dimension. Defaults to 'x'.
            x2 (str, optional): second dimension. Defaults to 'y'.
            x3_value (str, optional): third dimension. Defaults to None.
        """
        # Assign the axis names
        self.x1_name = x1
        self.x2_name = x2
        #TODO: if it will be assumed that x3 is one of x, y, or z that is x1 and x2,
        #       then we should verify that x1 and x2 are one of x, y, or z
        self.x3_name = [x3 for x3 in ['x', 'y', 'z'] if x3 not in [x1, x2]][0]

        # Get x1, x2 and x3 arrays
        x1_array = getattr(flow_data, self.x1_name)
        x2_array = getattr(flow_data, self.x2_name)
        x3_array = getattr(flow_data, self.x3_name)

        search_values = np.array(sorted(np.unique(x3_array)))
        nearest_idx = (np.abs(search_values - x3_value)).argmin()
        nearest_value = search_values[nearest_idx]
        print('Nearest value in %s to %.2f is %.2f' %
              (self.x3_name, x3_value, nearest_value))

        # Select down the data
        x3_select_mask = x3_array == nearest_value

        # Store the un-interpolated input arrays at this slice
        self.x1_in = x1_array[x3_select_mask]
        self.x2_in = x2_array[x3_select_mask]
        self.u_in = flow_data.u[x3_select_mask]
        self.v_in = flow_data.v[x3_select_mask]
        self.w_in = flow_data.w[x3_select_mask]

        # Initially, x1_lin, x2_lin are unique values of input
        self.x1_lin = np.unique(self.x1_in)
        self.x2_lin = np.unique(self.x2_in)

        # Save the resolution as the number of unique points in x1 and x2
        self.resolution = (len(np.unique(self.x1_lin)),
                           len(np.unique(self.x2_lin)))

        # Make initial meshing
        self._remesh()

    def _remesh(self):

        # Mesh and interpolate u, v and w
        self.x1_mesh, self.x2_mesh = np.meshgrid(self.x1_lin, self.x2_lin)
        self.u_mesh = griddata(
            np.column_stack([nudge_outward(self.x1_in), nudge_outward(self.x2_in)]),
            self.u_in, (self.x1_mesh.flatten(), self.x2_mesh.flatten()),
            method='cubic')
        self.v_mesh = griddata(
            np.column_stack([nudge_outward(self.x1_in), nudge_outward(self.x2_in)]),
            self.v_in, (self.x1_mesh.flatten(), self.x2_mesh.flatten()),
            method='cubic')
        self.w_mesh = griddata(
            np.column_stack([nudge_outward(self.x1_in), nudge_outward(self.x2_in)]),
            self.w_in, (self.x1_mesh.flatten(), self.x2_mesh.flatten()),
            method='cubic')

        # Save flat vectors
        self.x1_flat = self.x1_mesh.flatten()
        self.x2_flat = self.x2_mesh.flatten()

        # Save u-cubed
        self.u_cubed = self.u_mesh**3


# Define horizontal subclass
class HorPlane(_CutPlane):
    """
    Subclass of _CutPlane. Shortcut to extracting a horizontal plane.
    """

    def __init__(self, flow_data, z_value):
        """
        Initialize horizontal CutPlane

        Args:
            flow_data (np.array): 3D vector field of velocity data
            z_value (float): vertical position through which to slice
        """
        # Set up call super
        super().__init__(flow_data, x1='x', x2='y', x3_value=z_value)


# Define cross plane subclass
class CrossPlane(_CutPlane):
    """
    Subclass of _CutPlane. Shortcut to extracting a cross-stream plane.
    """

    def __init__(self, flow_data, x_value):
        """
        Initialize cross-stream CutPlane

        Args:
            flow_data (np.array): 3D vector field of velocity data
            x_value (float): streamwise position through which to slice
        """
        # Set up call super
        super().__init__(flow_data, x1='y', x2='z', x3_value=x_value)


# Define cross plane subclass
class VertPlane(_CutPlane):
    """
    Subclass of _CutPlane. Shortcut to extracting a streamwise-vertical plane.
    """

    def __init__(self, flow_data, y_value):
        """
        Initialize streamwise-vertical  CutPlane

        Args:
            flow_data (np.array): 3D vector field of velocity data
            y_value (float): spanwise position through which to slice
        """
        # Set up call super
        super().__init__(flow_data, x1='x', x2='z', x3_value=y_value)


## Modification functions
def set_origin(cut_plane, center_x1=0.0, center_x2=0.0):
    """
    Establish the origin of a CutPlane object.

    Args:
        cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`):
            plane of data.
        center_x1 (float, optional): x1-coordinate of orign.
                Defaults to 0.0.
        center_x2 (float, optional): x2-coordinate of orign.
                Defaults to 0.0.

    Returns:
        cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`):
                updated plane of data.
    """
    # Store the un-interpolated input arrays at this slice
    cut_plane.x1_in = cut_plane.x1_in - center_x1
    cut_plane.x2_in = cut_plane.x2_in - center_x2
    cut_plane.x1_lin = cut_plane.x1_lin - center_x1
    cut_plane.x2_lin = cut_plane.x2_lin - center_x2

    # Remesh
    cut_plane._remesh()

    return cut_plane


def change_resolution(cut_plane, resolution=(100, 100)):
    """
    Modify default resolution of a CutPlane object.

    Args:
        cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`):
            plane of data.
        resolution (tuple, optional): Desired resolution in x1 and x2.
            Defaults to (100, 100).

    Returns:
        cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`):
                updated plane of data.
    """
    # Grid the data
    cut_plane.x1_lin = np.linspace(min(cut_plane.x1_in), max(cut_plane.x1_in),
                                   resolution[0])
    cut_plane.x2_lin = np.linspace(min(cut_plane.x2_in), max(cut_plane.x2_in),
                                   resolution[1])

    # Save the new resolution
    cut_plane.resolution = resolution

    # Redo the mesh
    cut_plane._remesh()

    # Return the cutplane
    return cut_plane


def interpolate_onto_array(cut_plane, x1_array, x2_array):
    """
    Interpolate a CutPlane object onto specified coordinate arrays.

    Args:
        cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`):
            plane of data.
        x1_array (np.array): specified x1-coordinate.
        x2_array (np.array): specified x2-coordinate.

    Returns:
        cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`):
                updated plane of data.
    """
    # Grid the data given array
    cut_plane.x1_lin = x1_array
    cut_plane.x2_lin = x2_array

    # Save the new resolution
    cut_plane.resolution = (len(np.unique(cut_plane.x1_lin)),
                            len(np.unique(cut_plane.x2_lin)))

    # Redo the mesh
    cut_plane._remesh()

    # Return the cutplane
    return cut_plane


def rescale_axis(cut_plane, x1_factor=1.0, x2_factor=1.0):
    """
    Stretch or compress CutPlane coordinates.

    Args:
        cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`):
            plane of data.
        x1_factor (float): scaling factor for x1-coordinate.
        x2_factor (float): scaling factor for x2-coordinate.

    Returns:
        cut_plane (:py:class:`floris.tools.cut_plane._CutPlane`):
                updated plane of data.
    """
    # Store the un-interpolated input arrays at this slice
    cut_plane.x1_in = cut_plane.x1_in / x1_factor
    cut_plane.x2_in = cut_plane.x2_in / x2_factor
    cut_plane.x1_lin = cut_plane.x1_lin / x1_factor
    cut_plane.x2_lin = cut_plane.x2_lin / x2_factor

    # Remesh
    cut_plane._remesh()

    return cut_plane


def calculate_wind_speed(cross_plane, x1_loc, x2_loc, R):
    """
    Calculate effective wind speed within specified range of a point.

    Args:
        cross_plane (:py:class:`floris.tools.cut_plane.CrossPlane`):
            plane of data.
        x1_loc (float): x1-coordinate of point of interst.
        x2_loc (float): x2-coordinate of point of interst.
        R (float): radius from point of interst to consider

    Returns:
        (float): effective wind speed
    """
    # Make a distance column
    distance = np.sqrt((cross_plane.x1_flat - x1_loc)**2 +
                       (cross_plane.x2_flat - x2_loc)**2)

    # Return the mean wind speed
    return np.cbrt(np.mean(cross_plane.u_cubed[distance < R]))


def calculate_power(cross_plane,
                    x1_loc,
                    x2_loc,
                    R,
                    ws_array,
                    cp_array,
                    air_density=1.225):
    """
    Calculate maximum power available in a given cross plane.

    Args:
        cross_plane (:py:class:`floris.tools.cut_plane.CrossPlane`):
            plane of data.
        x1_loc (float): x1-coordinate of point of interst.
        x2_loc (float): x2-coordinate of point of interst.
        R (float): Radius of wind turbine rotor.
        ws_array (np.array): reference wind speed for cp curve.
        cp_array (np.array): cp curve at reference wind speeds.
        air_density (float, optional): air density. Defaults to 1.225.

    Returns:
        float: Power!
    """
    # Compute the ws
    ws = calculate_wind_speed(cross_plane, x1_loc, x2_loc, R)

    # Compute the cp
    cp_value = np.interp(ws, ws_array, cp_array)

    #Return the power
    return 0.5 * air_density * (np.pi * R**2) * cp_value * ws**3

    # def get_profile(self, R, x2_loc, resolution=100, x1_locs=None):
    #     if x1_locs is None:
    #         x1_locs = np.linspace(
    #             min(self.x1_flat), max(self.x1_flat), resolution)
    #     v_array = np.array([self.calculate_wind_speed(
    #         x1_loc, x2_loc, R) for x1_loc in x1_locs])
    #     return x1_locs, v_array)

    # def get_power_profile(self, ws_array, cp_array, rotor_radius, air_density=1.225, resolution=100, x1_locs=None):

    #     # Get the wind speed profile
    #     x1_locs, v_array = self.get_profile(resolution=resolution, x1_locs=x1_locs)

    #     # Get Cp
    #     cp_array = np.interp(v_array,ws_array,cp_array)

    #     # Return power array
    #     return x1_locs, 0.5 * air_density * (np.pi * rotor_radius**2) * cp_array * v_array**3



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

## ISSUES TO TACKLE
## 1: BINNING VALUES OUTSIDE OF WD AND WS (just ignore?)
## 2: Include smoothing?
## 3: Plotting
## 4: FLORIS return

import os
import dateutil
import h5pyd
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np
import pandas as pd
import pickle
from pyproj import Proj
import floris.utilities as geo


class WindRose():
    """
    WindRose object class used to parse data and generate figures.
    """

    def __init__(self, ):
        """
        Init Function of WindRose Object.
        No explicit arguments required.
        """

        # Initialize some varibles to zero
        self.num_wd = 0
        self.num_ws = 0
        self.wd_step = 1.
        self.ws_step = 5.
        self.wd = np.array([])
        self.ws = np.array([])
        self.df = pd.DataFrame()

    def save(self, filename):
        """
        Method for pickling WIndRose objects.

        Args:
            filename (str): Write-to path for WindRose pickle.
        """
        pickle.dump([
            self.num_wd, self.num_ws, self.wd_step, self.ws_step, self.wd,
            self.ws, self.df
        ], open(filename, "wb"))

    def load(self, filename):
        """
        Load previously pickled WindRose object.

        Args:
            filename (str): Read-from path for pickled WindRose Object
        """
        self.num_wd, self.num_ws, self.wd_step, self.ws_step, self.wd, self.ws, self.df = pickle.load(
            open(filename, "rb"))

    def resample_wind_speed(self, df, ws=np.arange(0, 26, 1.)):
        """
        Modify the default bins for sorting wind speed.

        Args:
            df (pd.DataFrame): Wind speed data
            ws (np.array, optional): Vector of wind speed bins for
                WindRose. Defaults to np.arange(0, 26, 1.).

        Returns:
            df (pd.DataFrame): Resampled wind speed for WindRose
        """
        # Make a copy of incoming dataframe
        df = df.copy(deep=True)

        # Get the wind step
        ws_step = ws[1] - ws[0]

        # Ws
        ws_edges = (ws - ws_step / 2.0)
        ws_edges = np.append(ws_edges, np.array(ws[-1] + ws_step / 2.0))

        # Cut wind speed onto bins
        df['ws'] = pd.cut(df.ws, ws_edges, labels=ws)

        # Regroup
        df = df.groupby(['ws', 'wd']).sum()

        # Fill nans
        df = df.fillna(0)

        # Reset the index
        df = df.reset_index()

        # Set to float
        df['ws'] = df.ws.astype(float)
        df['wd'] = df.wd.astype(float)

        return df

    def internal_resample_wind_speed(self, ws=np.arange(0, 26, 1.)):
        """
        Internal method for resampling wind speed into desired bins.
        Modifies data within WindRose object without explicit return.

        Args:
            ws (np.array, optional): Vector of wind speed bins for
                WindRose. Defaults to np.arange(0, 26, 1.).
        """
        # Update ws and wd binning
        self.ws = ws
        self.num_ws = len(ws)
        self.ws_step = ws[1] - ws[0]

        # Update internal data frame
        self.df = self.resample_wind_speed(self.df, ws)

    def resample_wind_direction(self, df, wd=np.arange(0, 360, 5.)):
        """
        Modify the default bins for sorting wind direction.

        Args:
            df (pd.DataFrame): Wind direction data
                wd (np.array, optional): Vector of wind direction bins
                for WindRose. Defaults to np.arange(0, 360, 5.).

        Returns:
            df (pd.DataFrame): Resampled wind direction for WindRose
        """
        # Make a copy of incoming dataframe
        df = df.copy(deep=True)

        # Get the wind step
        wd_step = wd[1] - wd[0]

        # Get bin edges
        wd_edges = (wd - wd_step / 2.0)
        wd_edges = np.append(wd_edges, np.array(wd[-1] + wd_step / 2.0))

        # Get the overhangs
        negative_overhang = wd_edges[0]
        positive_overhang = wd_edges[-1] - 360.

        # Need potentially to wrap high angle direction to negative for correct binning
        df['wd'] = geo.wrap_360(df.wd)
        if negative_overhang < 0:
            print('Correcting negative Overhang:%.1f' % negative_overhang)
            df['wd'] = np.where(df.wd.values >= 360. + negative_overhang,
                                df.wd.values - 360., df.wd.values)

        # Check on other side
        if positive_overhang > 0:
            print('Correcting positive Overhang:%.1f' % positive_overhang)
            df['wd'] = np.where(df.wd.values <= positive_overhang,
                                df.wd.values + 360., df.wd.values)

        # Cut into bins
        df['wd'] = pd.cut(df.wd, wd_edges, labels=wd)

        # Regroup
        df = df.groupby(['ws', 'wd']).sum()

        # Fill nans
        df = df.fillna(0)

        # Reset the index
        df = df.reset_index()

        # Set to float Re-wrap
        df['wd'] = df.wd.astype(float)
        df['ws'] = df.ws.astype(float)
        df['wd'] = geo.wrap_360(df.wd)

        return df

    def internal_resample_wind_direction(self, wd=np.arange(0, 360, 5.)):
        """
        Internal method for resampling wind direction into desired bins.
        Modifies data within WindRose object without explicit return.

        Args:
            wd (np.array, optional): Vector of wind direction bins for
                WindRose. Defaults to np.arange(0, 360, 5.).
        """
        # Update ws and wd binning
        self.wd = wd
        self.num_wd = len(wd)
        self.wd_step = wd[1] - wd[0]

        # Update internal data frame
        self.df = self.resample_wind_direction(self.df, wd)

    def resample_average_ws_by_wd(self, df):
        """
        Re-established counts of wind speed observations in wind
        direction bins.

        Args:
            df (pd.DataFrame): Wind speed and direction data

        Returns:
            df (pd.DataFrame): Resampled wind speed and direction data.
        """
        # Make a copy of incoming dataframe
        df = df.copy(deep=True)

        ws_avg = []

        for val in df.wd.unique():
            ws_avg.append(
                np.array(df.loc[df['wd'] == val]['ws'] *
                         df.loc[df['wd'] == val]['freq_val']).sum() /
                df.loc[df['wd'] == val]['freq_val'].sum())

        # Regroup
        df = df.groupby('wd').sum()

        df['ws'] = ws_avg

        # Reset the index
        df = df.reset_index()

        # Set to float
        df['ws'] = df.ws.astype(float)
        df['wd'] = df.wd.astype(float)

        return df

    def internal_resample_average_ws_by_wd(self, wd=np.arange(0, 360, 5.)):
        """
        Internal method for re-established counts of wind speed
        observations in wind direction bins.
        Modifies data within WindRose Object without explicit return.

        Args:
            wd (np.arange, optional): Wind direction bins limists.
                Defaults to np.arange(0, 360, 5.).
        """
        # Update ws and wd binning
        self.wd = wd
        self.num_wd = len(wd)
        self.wd_step = wd[1] - wd[0]
        # self #TODO This reference to 'self' was here alone. Is this intentional as a print? is it incomplete?

        # Update internal data frame
        self.df = self.resample_average_ws_by_wd(self.df)

    def weibull(self, x, k=2.5, lam=8.0):
        """
        Weibull distribution function object for least-squares fitting.

        Args:
            x (np.array): Input data (typically binned wind speed
                observations.)
            k (float, optional): Weibull share parameter.
                Defaults to 2.5.
            lam (float, optional): Weibull scale parameter.
                Defaults to 8.0.

        Returns:
            np.array: Weibull distribution
        """
        return (k / lam) * (x / lam)**(k - 1) * np.exp(-(x / lam)**k)

    def make_wind_rose_from_weibull(self,
                                    wd=np.arange(0, 360, 5.),
                                    ws=np.arange(0, 26, 1.)):
        """
        Populate binned observations of wind speed and direction from
        fitted Weibull distribution.

        Args:
            wd (np.array, optional): Wind direciton bins.
                Defaults to np.arange(0, 360, 5.).
            ws (np.array, optional): Wind speed bins.
                Defaults to np.arange(0, 26, 1.).

        Returns:
            df (pd.DataFrame): updated wind speed and direction bins.
            #TODO should these be returned or updated internally, both?
        """
        # Use an assumed wind-direction for dir frequency
        wind_dir = [
            0, 22.5, 45, 67.5, 90, 112.5, 135, 157.5, 180, 202.5, 225, 247.5,
            270, 292.5, 315, 337.5
        ]
        freq_dir = [
            0.064, 0.04, 0.038, 0.036, 0.045, 0.05, 0.07, 0.08, 0.11, 0.08,
            0.05, 0.036, 0.048, 0.058, 0.095, 0.10
        ]

        freq_wd = np.interp(wd, wind_dir, freq_dir)
        freq_ws = self.weibull(ws)

        freq_tot = np.zeros(len(wd) * len(ws))
        wd_tot = np.zeros(len(wd) * len(ws))
        ws_tot = np.zeros(len(wd) * len(ws))

        count = 0
        for i in range(len(wd)):
            for j in range(len(ws)):
                wd_tot[count] = wd[i]
                ws_tot[count] = ws[j]

                freq_tot[count] = freq_wd[i] * freq_ws[j]
                count = count + 1

        # renormalize
        freq_tot = freq_tot / np.sum(freq_tot)

        # Load the wind toolkit data into a dataframe
        df = pd.DataFrame()

        # Start by simply round and wrapping the wind direction and wind speed columns
        df['wd'] = wd_tot
        df['ws'] = ws_tot

        # Now group up
        df['freq_val'] = freq_tot

        # Save the df at this point
        self.df = df
        #TODO is there a reason self.df is updated AND returned?
        return self.df

    def parse_wind_toolkit_folder(self,
                                  folder_name,
                                  wd=np.arange(0, 360, 5.),
                                  ws=np.arange(0, 26, 1.),
                                  limit_month=None):
        """
        Load and

        Args:
            folder_name (str): path to wind toolkit input data.
            wd (np.array, optional): Wind direction bin limits.
                Defaults to np.arange(0, 360, 5.).
            ws (np.array, optional): Wind speed bin limits.
                Defaults to np.arange(0, 26, 1.).
            limit_month (str, optional): name of month(s) to consider.
                Defaults to None.

        Returns:
            df (pd.DataFrame): DataFrame updated with wind toolkit info.
        """
        # Load the wind toolkit data into a dataframe
        df = self.load_wind_toolkit_folder(folder_name,
                                           limit_month=limit_month)

        # Start by simply round and wrapping the wind direction and wind speed columns
        df['wd'] = geo.wrap_360(df.wd.round())
        df['ws'] = geo.wrap_360(df.ws.round())

        # Now group up
        df['freq_val'] = 1.
        df = df.groupby(['ws', 'wd']).sum()
        df['freq_val'] = df.freq_val.astype(float) / df.freq_val.sum()
        df = df.reset_index()

        # Save the df at this point
        self.df = df

        # Resample onto the provided wind speed and wind direction binnings
        self.internal_resample_wind_speed(ws=ws)
        self.internal_resample_wind_direction(wd=wd)

        return self.df

    def load_wind_toolkit_folder(self, folder_name, limit_month=None):
        """
        Given a wind_toolkit folder of files, produce a list of
        files to input

        Args:
            folder_name (str): path to folder containing wind toolikit
                data
            limit_month (str, optional): limit loaded data to specified
                month. Defaults to None.

        Returns:
            df (pd.DataFrame): DataFrame containing wind data from each
                of the files in the folder.
        """

        file_list = os.listdir(folder_name)
        file_list = [
            os.path.join(folder_name, f) for f in file_list if '.csv' in f
        ]

        df = pd.DataFrame()
        for f_idx, f in enumerate(file_list):
            print('%d of %d: %s' % (f_idx, len(file_list), f))
            df_temp = self.load_wind_toolkit_file(f, limit_month=limit_month)
            df = df.append(df_temp)

        return df

    def load_wind_toolkit_file(self, filename, limit_month=None):
        """
        Load a particular wind toolkit data file.

        Args:
            filename (str): path to data file.
            limit_month (str, optional): limit loaded data to specified
                month. Defaults to None.

        Returns:
            df (pd.DataFrame): DataFrame containing wind data from the
                specified file.
        """
        df = pd.read_csv(filename, header=3, sep=',')

        # If asked to limit to particular months
        if not limit_month is None:
            df = df[df.Month.isin(limit_month)]

        # Save just what I want
        speed_column = [c for c in df.columns if 'speed' in c][0]
        direction_column = [c for c in df.columns if 'direction' in c][0]
        df = df.rename(index=str,
                       columns={
                           speed_column: "ws",
                           direction_column: "wd"
                       })[['wd', 'ws']]

        return df

    def import_from_wind_toolkit_hsds(self,
                                      lat,
                                      lon,
                                      ht=100,
                                      wd=np.arange(0, 360, 5.),
                                      ws=np.arange(0, 26, 1.),
                                      limit_month=None,
                                      st_date=None,
                                      en_date=None):
        """
        Given a lat/long coordinate in the continental US return a
        dataframe containing the normalized frequency of each pair of
        wind speed and wind direction values specified. The wind data
        is obtained from the WIND Toolkit dataset (https://www.nrel.gov/
        grid/wind-toolkit.html) using the HSDS service (see https://
        github.com/NREL/hsds-examples). The wind data returned is
        obtained from the nearest 2km x 2km grid point to the input
        coordinate and is limited to the years 2007-2013.

        Requires h5pyd package, which can be installed using:
            pip install --user git+http://github.com/HDFGroup/h5pyd.git

        Then, make a configuration file at ~/.hscfg containing:

            hs_endpoint = https://developer.nrel.gov/api/hsds/
            hs_username = None
            hs_password = None
            hs_api_key = 3K3JQbjZmWctY0xmIfSYvYgtIcM3CN0cb1Y2w9bf

        The example API key above is for demonstation and is
        rate-limited per IP. To get your own API key, visit https://
        developer.nrel.gov/signup/

        More information can be found at: https://github.com/NREL/
        hsds-examples

        Args:
            lat (float): coordinates in degrees
            lon (float): coordinates in degrees
            ht (int, optional): height above ground where wind
                information is obtained (m). Defaults to 100.
            wd (np.array, optional): Wind direction bin limits.
                Defaults to np.arange(0, 360, 5.).
            ws (np.array, optional): Wind speed bin limits.
                Defaults to np.arange(0, 26, 1.).
            limit_month (str, optional): limit loaded data to specified
                month. Defaults to None.
            st_date (str, optional): 'MM-DD-YYYY'.
                Defaults to None.
            en_date (str, optional): 'MM-DD-YYYY'.
                Defaults to None.

        Returns:
            df (pd.DataFrame): DataFrame with wind speed and direction
                data.
        """

        # check inputs
        if st_date is not None:
            if dateutil.parser.parse(st_date) > dateutil.parser.parse(
                    '12-13-2013 23:00'):
                print(
                    'Error, invalid date range. Valid range: 01-01-2007 - 12/31/2013'
                )
                return None
        if en_date is not None:
            if dateutil.parser.parse(en_date) < dateutil.parser.parse(
                    '01-01-2007 00:00'):
                print(
                    'Error, invalid date range. Valid range: 01-01-2007 - 12/31/2013'
                )
                return None

        if ht not in [10, 40, 60, 80, 100, 120, 140, 160, 200]:
            print(
                'Error, invalid height. Valid heights: 10, 40, 60, 80, 100, 120, 140, 160, 200'
            )
            return None

        # load wind speeds and directions from WIND Toolkit
        df = self.load_wind_toolkit_hsds(lat,
                                         lon,
                                         ht=ht,
                                         limit_month=limit_month,
                                         st_date=st_date,
                                         en_date=en_date)

        # check for errors in loading wind toolkit data
        if df is None:
            return None

        # Start by simply round and wrapping the wind direction and wind speed columns
        df['wd'] = geo.wrap_360(df.wd.round())
        df['ws'] = geo.wrap_360(df.ws.round())

        # Now group up
        df['freq_val'] = 1.
        df = df.groupby(['ws', 'wd']).sum()
        df['freq_val'] = df.freq_val.astype(float) / df.freq_val.sum()
        df = df.reset_index()

        # Save the df at this point
        self.df = df

        # Resample onto the provided wind speed and wind direction binnings
        self.internal_resample_wind_speed(ws=ws)
        self.internal_resample_wind_direction(wd=wd)

        return df

    def load_wind_toolkit_hsds(self,
                               lat,
                               lon,
                               ht=100,
                               limit_month=None,
                               st_date=None,
                               en_date=None):
        """
        Given a lat/long coordinate in the continental US return a
        dataframe containing wind speeds and wind directions at the
        specified height for each hour in the date range given. The
        wind data is obtained from the WIND Toolkit dataset (https://
        www.nrel.gov/grid/wind-toolkit.html) using the HSDS service
        (see https://github.com/NREL/hsds-examples). The wind data
        returned is obtained from the nearest 2km x 2km grid point to
        the input coordinate.

        Args:
            lat (float): coordinates in degrees.
            lon (float): coordinates in degrees.
            ht (int, optional): height above ground where wind
                information is obtained (m).
                Defaults to 100.
            limit_month (str, optional): limit loaded data to specified
                month. Defaults to None.
            st_date (str, optional): 'MM-DD-YYYY'.
                Defaults to None.
            en_date (str, optional): 'MM-DD-YYYY'.
                Defaults to None.

        Returns:
            df (pd.DataFrame): dataframe with wind speed and wind
                directiond columns containing hourly data.
        """

        # Open the wind data "file"
        # server endpoint, username, password is found via a config file
        f = h5pyd.File("/nrel/wtk-us.h5", 'r')

        # assign wind direction, wind speed, and time datasets for the desired height
        wd_dset = f['winddirection_' + str(ht) + 'm']
        ws_dset = f['windspeed_' + str(ht) + 'm']
        dt = f['datetime']
        dt = pd.DataFrame({'datetime': dt[:]}, index=range(0, dt.shape[0]))
        dt['datetime'] = dt['datetime'].apply(dateutil.parser.parse)

        # find dataset indices from lat/long
        Location_idx = self.indices_for_coord(f, lat, lon)

        # check if in bounds
        if (Location_idx[0] < 0) | (Location_idx[0] >= wd_dset.shape[1]) | (
                Location_idx[1] < 0) | (Location_idx[1] >= wd_dset.shape[2]):
            print(
                'Error, coordinates out of bounds. WIND Toolkit database covers the continental United States.'
            )
            return None

        # create dataframe with wind direction and wind speed
        df = pd.DataFrame()
        df['wd'] = wd_dset[:, Location_idx[0], Location_idx[1]]
        df['ws'] = ws_dset[:, Location_idx[0], Location_idx[1]]
        df['datetime'] = dt['datetime']

        # limit dates if start and end dates are provided
        if (st_date is not None):
            df = df[df.datetime >= st_date]

        if (en_date is not None):
            df = df[df.datetime < en_date]

        # limit to certain months if specified
        if not limit_month is None:
            df['month'] = df['datetime'].map(lambda x: x.month)
            df = df[df.month.isin(limit_month)]
        df = df[['wd', 'ws']]

        return df

    def indices_for_coord(self, f, lat_index, lon_index):
        #TODO This function is tough for me to follow. What is f?
        """
        Function obtained from: https://github.com/NREL/hsds-examples/
        blob/master/notebooks/01_introduction.ipynb "indicesForCoord"
        This function finds the nearest x/y indices for a given lat/lon.
        Rather than fetching the entire coordinates database, which is
        500+ MB, this uses the Proj4 library to find a nearby point and
        then converts to x/y indices.

        Args:
            f (dict): #TODO [description]
            lat_index (int): index to desired latitude coordinate.
            lon_index (int): index to desired longitude coordinate.

        Returns:
            tuple: #TODO [description]
        """

        dset_coords = f['coordinates']
        projstring = """+proj=lcc +lat_1=30 +lat_2=60
                        +lat_0=38.47240422490422 +lon_0=-96.0
                        +x_0=0 +y_0=0 +ellps=sphere
                        +units=m +no_defs """
        projectLcc = Proj(projstring)
        origin_ll = reversed(
            dset_coords[0][0])  # Grab origin directly from database
        origin = projectLcc(*origin_ll)

        coords = (lon_index, lat_index)
        coords = projectLcc(*coords)
        delta = np.subtract(coords, origin)
        ij = [int(round(x / 2000)) for x in delta]
        return tuple(reversed(ij))

    def plot_wind_speed_all(self, ax=None):
        """
        Add binned wind speed observations to a plot. If no axis is
        provided, make a new one.

        Args:
            ax (:py:class:`matplotlib.pyplot.axes`, optional): Figure
                axes to which data should be plotted. Defaults to None.
        """
        if ax is None:
            _, ax = plt.subplots()

        df_plot = self.df.groupby('ws').sum()
        ax.plot(self.ws, df_plot.freq_val)

    def plot_wind_speed_by_direction(self, dirs, ax=None):
        """
        Add  wind speed observations binned by wind direction to a
        plot. If no axis is provided, make a new one.

        Args:
            dirs (np.array): vector of wind direction bins.
            ax (:py:class:`matplotlib.pyplot.axes`, optional): Figure
                axes to which data should be plotted. Defaults to None.
        """
        # Get a downsampled frame
        df_plot = self.resample_wind_direction(self.df, wd=dirs)

        if ax is None:
            _, ax = plt.subplots()

        for wd in dirs:
            df_plot_sub = df_plot[df_plot.wd == wd]
            ax.plot(df_plot_sub.ws, df_plot_sub['freq_val'], label=wd)
        ax.legend()

    def plot_wind_rose(self,
                       ax=None,
                       color_map='viridis_r',
                       ws_right_edges=np.array([5, 10, 15, 20, 25]),
                       wd_bins=np.arange(0, 360, 15.)):
        """
        Generate wind rose plot. If no axis is provided, make a new one.

        Args:
            ax (:py:class:`matplotlib.pyplot.axes`, optional): Figure axes to which data should
                be plotted. Defaults to None.
            color_map (str, optional): name of colormap.
                Defaults to 'viridis_r'.
            ws_right_edges (np.array, optional): upper bounds of wind
                speed bins. Defaults to np.array([5, 10, 15, 20, 25]).
            wd_bins (np.array, optional): wind direction bin limits.
                Defaults to np.arange(0, 360, 15.).

        Returns:
            ax (:py:class:`matplotlib.pyplot.axes`): Figure axes
                containing wind rose plot.
        """
        # Based on code provided by Patrick Murphy

        # Resample data onto bins
        # df_plot = self.resample_wind_speed(self.df,ws=ws_bins)
        df_plot = self.resample_wind_direction(self.df, wd=wd_bins)

        # Make labels for wind speed based on edges
        ws_step = ws_right_edges[1] - ws_right_edges[0]
        # ws = ws_edges
        # ws_edges = (ws - ws_step / 2.0)
        # ws_edges = np.append(ws_edges,np.array(ws[-1] + ws_step / 2.0))
        # ws_edges = np.append([0], ws_right_edges)
        ws_labels = ['%d-%d m/s' % (w - ws_step, w) for w in ws_right_edges]

        # Grab the wd_step
        wd_step = wd_bins[1] - wd_bins[0]

        # Set up figure
        if ax is None:
            _, ax = plt.subplots(subplot_kw=dict(polar=True))

        # Get a color array
        color_array = cm.get_cmap(color_map, len(ws_right_edges))

        for wd_idx, wd in enumerate(wd_bins):
            rects = list()
            df_plot_sub = df_plot[df_plot.wd == wd]
            for ws_idx, ws in enumerate(ws_right_edges[::-1]):
                plot_val = df_plot_sub[df_plot_sub.ws <= ws].freq_val.sum(
                )  # Get the sum of frequency up to this wind speed
                rects.append(
                    ax.bar(np.radians(wd),
                           plot_val,
                           width=0.9 * np.radians(wd_step),
                           color=color_array(ws_idx),
                           edgecolor='k'))
            # break

        # Configure the plot
        ax.legend(reversed(rects), ws_labels)
        ax.set_theta_direction(-1)
        ax.set_theta_offset(np.pi / 2.0)
        ax.set_theta_zero_location("N")
        ax.set_xticklabels(['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW'])

        return ax

    def export_for_floris_opt(self):
        """
        Shortcut function to generate output for FLORIS.

        Returns:
            (list): tuples containing DataFrame values.
        """
        # Return a list of tuples, where each tuple is (ws,wd,freq)
        return [tuple(x) for x in self.df.values]


# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""
The :py:obj:`floris.simulation` package contains the modules used to
complete a wake simulation for a given wind farm and turbine
configuration.

All classes can be be imported with

    >>> import floris.simulation

The ``__init__.py`` file enables the import of all modules in this
package so any additional modules should be included there.

Examples:
    >>> import floris.simulation
    
    >>> dir(floris.simulation)
    ['Farm', 'Floris', 'FlowField', 'InputReader', 'Turbine',
    'TurbineMap', 'Wake', 'WakeCombination', 'WakeDeflection',
    'WakeVelocity', '__builtins__', '__cached__', '__doc__',
    '__file__', '__loader__', '__name__', '__package__', '__path__',
    '__spec__', 'farm', 'floris', 'flow_field', 'input_reader',
    'turbine', 'turbine_map', 'wake', 'wake_combination',
    'wake_deflection', 'wake_velocity']
"""

from .farm import Farm
from .floris import Floris
from .flow_field import FlowField
from .input_reader import InputReader
from .turbine_map import TurbineMap
from .turbine import Turbine
from .wake_combination import WakeCombination
from .wake_deflection import WakeDeflection
from .wake_velocity import WakeVelocity
from .wake import Wake



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import numpy as np
from scipy.ndimage.filters import gaussian_filter
from ..utilities import Vec3
from ..utilities import cosd, sind, tand
import copy


class WakeVelocity():
    """
    WakeVelocity is the base class of the different wake velocity model 
    classes.

    An instantiated WakeVelocity object will import parameters used to 
    calculate wake-added turbulence intensity from an upstream turbine, 
    using the approach of Crespo, A. and Herna, J., "Turbulence 
    characteristics in wind-turbine wakes." *J. Wind Eng Ind Aerodyn*. 
    1996.

    Args:
        parameter_dictionary: A dictionary as generated from the 
            input_reader; it should have the following key-value pairs:

            -   **turbulence_intensity**: A dictionary containing the 
                following key-value pairs:

                -   **initial**: A float that is the initial ambient 
                    turbulence intensity, expressed as a decimal 
                    fraction.
                -   **constant**: A float that is the constant used to 
                    scale the wake-added turbulence intensity.
                -   **ai**: A float that is the axial induction factor 
                    exponent used in in the calculation of wake-added 
                    turbulence.
                -   **downstream**: A float that is the exponent 
                    applied to the distance downtream of an upstream 
                    turbine normalized by the rotor diameter used in 
                    the calculation of wake-added turbulence.

    Returns:
        An instantiated WakeVelocity object.
    """

    def __init__(self, parameter_dictionary):
        self.requires_resolution = False
        self.model_string = None
        self.model_grid_resolution = None

        # turbulence parameters
        turbulence_intensity = parameter_dictionary["turbulence_intensity"]
        self.ti_initial = float(turbulence_intensity["initial"])
        self.ti_constant = float(turbulence_intensity["constant"])
        self.ti_ai = float(turbulence_intensity["ai"])
        self.ti_downstream = float(turbulence_intensity["downstream"])

    def __str__(self):
        return self.model_string


class Jensen(WakeVelocity):
    """
    Wake velocity deficit model based on the Jensen model.

    Jensen is a subclass of :py:class:`floris.simulation.wake_velocity.WakeVelocity` that is 
    used to compute the wake velocity deficit based on the classic 
    Jensen/Park model. See Jensen, N. O., "A note on wind generator 
    interaction." Tech. Rep. Risø-M-2411, Risø, 1983.

    Args:
        parameter_dictionary: A dictionary as generated from the 
            input_reader; it should have the following key-value pairs:

            -   **turbulence_intensity**: A dictionary containing the 
                following key-value pairs used to calculate wake-added 
                turbulence intensity from an upstream turbine, using 
                the approach of Crespo, A. and Herna, J. "Turbulence 
                characteristics in wind-turbine wakes." *J. Wind Eng 
                Ind Aerodyn*. 1996.:

                -   **initial**: A float that is the initial ambient 
                    turbulence intensity, expressed as a decimal 
                    fraction.
                -   **constant**: A float that is the constant used to 
                    scale the wake-added turbulence intensity.
                -   **ai**: A float that is the axial induction factor 
                    exponent used in in the calculation of wake-added 
                    turbulence.
                -   **downstream**: A float that is the exponent 
                    applied to the distance downtream of an upstream 
                    turbine normalized by the rotor diameter used in 
                    the calculation of wake-added turbulence.

            -   **jensen**: A dictionary containing the following 
                key-value pairs:

                -   **we**: A float that is the linear wake decay 
                    constant that defines the cone boundary for the 
                    wake as well as the velocity deficit. D/2 +/- we*x 
                    is the cone boundary for the wake.

    Returns:
        An instantiated Jensen object.
    """

    def __init__(self, parameter_dictionary):
        super().__init__(parameter_dictionary)
        self.model_string = "jensen"
        model_dictionary = parameter_dictionary[self.model_string]
        self.we = float(model_dictionary["we"])

    def function(self, x_locations, y_locations, z_locations, turbine, turbine_coord, deflection_field, wake, flow_field):
        """
        Using the Jensen wake model, this method calculates and returns 
        the wake velocity deficits, caused by the specified turbine, 
        relative to the freestream velocities at the grid of points 
        comprising the wind farm flow field.

        Args:
            x_locations: An array of floats that contains the 
                streamwise direction grid coordinates of the flow field 
                domain (m).
            y_locations: An array of floats that contains the grid 
                coordinates of the flow field domain in the direction 
                normal to x and parallel to the ground (m).
            z_locations: An array of floats that contains the grid 
                coordinates of the flow field domain in the vertical 
                direction (m).
            turbine: A :py:obj:`floris.simulation.turbine` object that 
                represents the turbine creating the wake.
            turbine_coord: A :py:obj:`floris.utilities.Vec3` object 
                containing the coordinate of the turbine creating the 
                wake (m).
            deflection_field: An array of floats that contains the 
                amount of wake deflection in meters in the y direction 
                at each grid point of the flow field.
            wake: A :py:obj:`floris.simulation.wake` object containing 
                the wake model used.
            flow_field: A :py:class:`floris.simulation.flow_field` 
                object containing the flow field information for the 
                wind farm.

        Returns:
            Three arrays of floats that contain the wake velocity 
            deficit in m/s created by the turbine relative to the 
            freestream velocities for the u, v, and w components, 
            aligned with the x, y, and z directions, respectively. The 
            three arrays contain the velocity deficits at each grid 
            point in the flow field. 
        """

        # define the boundary of the wake model ... y = mx + b
        m = self.we
        x = x_locations - turbine_coord.x1
        b = turbine.rotor_radius

        boundary_line = m * x + b

        y_upper = boundary_line + turbine_coord.x2 + deflection_field
        y_lower = -1 * boundary_line + turbine_coord.x2 + deflection_field

        z_upper = boundary_line + turbine.hub_height
        z_lower = -1 * boundary_line + turbine.hub_height

        # calculate the wake velocity
        c = (turbine.rotor_diameter /
             (2 * self.we * (x_locations - turbine_coord.x1) + turbine.rotor_diameter))**2

        # filter points upstream and beyond the upper and lower bounds of the wake
        c[x_locations - turbine_coord.x1 < 0] = 0
        c[y_locations > y_upper] = 0
        c[y_locations < y_lower] = 0
        c[z_locations > z_upper] = 0
        c[z_locations < z_lower] = 0

        return 2 * turbine.aI * c * flow_field.u_initial, np.zeros(np.shape(flow_field.u_initial)), np.zeros(np.shape(flow_field.u_initial))


class MultiZone(WakeVelocity):
    """
    Floris is a subclass of 
    :py:class:`floris.simulation.wake_velocity.WakeVelocity` that is 
    used to compute the wake velocity deficit based on the original 
    multi-zone FLORIS model. See: 

    Gebraad, P. M. O. et al., "A Data-driven model for wind plant power 
    optimization by yaw control." *Proc. American Control Conference*, 
    Portland, OR, 2014.

    Gebraad, P. M. O. et al., "Wind plant power optimization through 
    yaw control using a parametric model for wake effects - a CFD 
    simulation study." *Wind Energy*, 2016.

    Args:
        parameter_dictionary: A dictionary as generated from the 
            input_reader; it should have the following key-value pairs:

            -   **turbulence_intensity**: A dictionary containing the 
                following key-value pairs used to calculate wake-added 
                turbulence intensity from an upstream - turbine, using 
                the approach of Crespo, A. and Herna, J. "Turbulence 
                characteristics in wind-turbine wakes." *J. Wind Eng 
                Ind Aerodyn*. 1996.:

                -   **initial**: A float that is the initial ambient 
                    turbulence intensity, expressed as a decimal 
                    fraction.
                -   **constant**: A float that is the constant used to 
                    scale the wake-added turbulence intensity.
                -   **ai**: A float that is the axial induction factor 
                    exponent used in in the calculation of wake-added 
                    turbulence.
                -   **downstream**: A float that is the exponent 
                    applied to the distance downtream of an upstream 
                    turbine normalized by the rotor diameter used in 
                    the calculation of wake-added turbulence.

            - **floris**: A dictionary containing the following 
                key-value pairs:

                -   **me**: A list of three floats that help determine 
                    the slope of the diameters of the three wake zones 
                    (near wake, far wake, mixing zone) as a function of 
                    downstream distance.
                -   **we**: A float that is the scaling parameter used 
                    to adjust the wake expansion, helping to determine 
                    the slope of the diameters of the three wake zones 
                    as a function of downstream distance, as well as 
                    the recovery of the velocity deficits in the wake 
                    as a function of downstream distance.
                -   **aU**: A float that is a parameter used to 
                    determine the dependence of the wake velocity 
                    deficit decay rate on the rotor yaw angle.
                -   **bU**: A float that is another parameter used to 
                    determine the dependence of the wake velocity 
                    deficit decay rate on the rotor yaw angle.
                -   **mU**: A list of three floats that are parameters 
                    used to determine the dependence of the wake 
                    velocity deficit decay rate for each of the three 
                    wake zones on the rotor yaw angle.

    Returns:
        An instantiated Floris object.
    """

    def __init__(self, parameter_dictionary):
        super().__init__(parameter_dictionary)
        self.model_string = "multizone"
        model_dictionary = parameter_dictionary[self.model_string]
        self.me = [float(n) for n in model_dictionary["me"]]
        self.we = float(model_dictionary["we"])
        self.aU = float(model_dictionary["aU"])
        self.bU = float(model_dictionary["bU"])
        self.mU = [float(n) for n in model_dictionary["mU"]]

    def function(self, x_locations, y_locations, z_locations, turbine, turbine_coord, deflection_field, wake, flow_field):
        """
        Using the original FLORIS multi-zone wake model, this method 
        calculates and returns the wake velocity deficits, caused by 
        the specified turbine, relative to the freestream velocities at 
        the grid of points comprising the wind farm flow field.

        Args:
            x_locations: An array of floats that contains the 
                streamwise direction grid coordinates of the flow field 
                domain (m).
            y_locations: An array of floats that contains the grid 
                coordinates of the flow field domain in the direction 
                normal to x and parallel to the ground (m).
            z_locations: An array of floats that contains the grid 
                coordinates of the flow field domain in the vertical 
                direction (m).
            turbine: A :py:obj:`floris.simulation.turbine` object that 
                represents the turbine creating the wake.
            turbine_coord: A :py:obj:`floris.utilities.Vec3` object 
                containing the coordinate of the turbine creating the 
                wake (m).
            deflection_field: An array of floats that contains the 
                amount of wake deflection in meters in the y direction 
                at each grid point of the flow field.
            wake: A :py:obj:`floris.simulation.wake` object containing 
                the wake model used.
            flow_field: A :py:class:`floris.simulation.flow_field` 
                object containing the flow field information for the 
                wind farm.

        Returns:
            Three arrays of floats that contain the wake velocity 
            deficit in m/s created by the turbine relative to the 
            freestream velocities for the u, v, and w components, 
            aligned with the x, y, and z directions, respectively. The 
            three arrays contain the velocity deficits at each grid 
            point in the flow field. 
        """

        mu = self.mU / cosd(self.aU + self.bU * turbine.yaw_angle)

        # distance from wake centerline
        rY = abs(y_locations - (turbine_coord.x2 + deflection_field))
        # rZ = abs(z_locations - (turbine.hub_height))
        dx = x_locations - turbine_coord.x1

        # wake zone diameters
        nearwake = turbine.rotor_radius + self.we * self.me[0] * dx
        farwake = turbine.rotor_radius + self.we * self.me[1] * dx
        mixing = turbine.rotor_radius + self.we * self.me[2] * dx

        # initialize the wake field
        c = np.zeros(x_locations.shape)

        # near wake zone
        mask = rY <= nearwake
        c += mask * (turbine.rotor_diameter /
                     (turbine.rotor_diameter + 2 * self.we * mu[0] * dx))**2
        #mask = rZ <= nearwake
        #c += mask * (radius / (radius + we * mu[0] * dx))**2

        # far wake zone
        # ^ is XOR, x^y:
        #   Each bit of the output is the same as the corresponding bit in x
        #   if that bit in y is 0, and it's the complement of the bit in x
        #   if that bit in y is 1.
        # The resulting mask is all the points in far wake zone that are not
        # in the near wake zone
        mask = (rY <= farwake) ^ (rY <= nearwake)
        c += mask * (turbine.rotor_diameter /
                     (turbine.rotor_diameter + 2 * self.we * mu[1] * dx))**2
        #mask = (rZ <= farwake) ^ (rZ <= nearwake)
        #c += mask * (radius / (radius + we * mu[1] * dx))**2

        # mixing zone
        # | is OR, x|y:
        #   Each bit of the output is 0 if the corresponding bit of x AND
        #   of y is 0, otherwise it's 1.
        # The resulting mask is all the points in mixing zone that are not
        # in the far wake zone and not in  near wake zone
        mask = (rY <= mixing) ^ ((rY <= farwake) | (rY <= nearwake))
        c += mask * (turbine.rotor_diameter /
                     (turbine.rotor_diameter + 2 * self.we * mu[2] * dx))**2
        #mask = (rZ <= mixing) ^ ((rZ <= farwake) | (rZ <= nearwake))
        #c += mask * (radius / (radius + we * mu[2] * dx))**2

        # filter points upstream
        c[x_locations - turbine_coord.x1 < 0] = 0

        return 2 * turbine.aI * c * flow_field.wind_speed, np.zeros(np.shape(c)), np.zeros(np.shape(c))


class Gauss(WakeVelocity):
    """
    Gauss is a wake velocity subclass that contains objects related to the 
    Gaussian wake model.

    Gauss is a subclass of 
    :py:class:`floris.simulation.wake_velocity.WakeVelocity` that is 
    used to compute the wake velocity deficit based on the Gaussian 
    wake model with self-similarity. The Gaussian wake model includes a 
    Gaussian wake velocity deficit profile in the y and z directions 
    and includes the effects of ambient turbulence, added turbulence 
    from upstream wakes, as well as wind shear and wind veer. For more 
    information about the Gauss wake model theory, see: 

    Abkar, M. and Porte-Agel, F. "Influence of atmospheric stability on 
    wind-turbine wakes: A large-eddy simulation study." *Physics of 
    Fluids*, 2015. 

    Bastankhah, M. and Porte-Agel, F. "A new analytical model for 
    wind-turbine wakes." *Renewable Energy*, 2014.

    Bastankhah, M. and Porte-Agel, F. "Experimental and theoretical 
    study of wind turbine wakes in yawed conditions." *J. Fluid 
    Mechanics*, 2016.

    Niayifar, A. and Porte-Agel, F. "Analytical modeling of wind farms: 
    A new approach for power prediction." *Energies*, 2016.

    Dilip, D. and Porte-Agel, F. "Wind turbine wake mitigation through 
    blade pitch offset." *Energies*, 2017.

    Args:
        parameter_dictionary: A dictionary as generated from the 
            input_reader; it should have the following key-value pairs:

            -   **turbulence_intensity**: A dictionary containing the 
                following key-value pairs used to calculate wake-added 
                turbulence intensity from an upstream turbine, using 
                the approach of Crespo, A. and Herna, J. "Turbulence 
                characteristics in wind-turbine wakes." *J. Wind Eng 
                Ind Aerodyn*. 1996.:

                -   **initial**: A float that is the initial ambient 
                    turbulence intensity, expressed as a decimal 
                    fraction.
                -   **constant**: A float that is the constant used to 
                    scale the wake-added turbulence intensity.
                -   **ai**: A float that is the axial induction factor 
                    exponent used in in the calculation of wake-added 
                    turbulence.
                -   **downstream**: A float that is the exponent 
                    applied to the distance downtream of an upstream 
                    turbine normalized by the rotor diameter used in 
                    the calculation of wake-added turbulence.

            -   **gauss**: A dictionary containing the following 
                key-value pairs:

                -   **ka**: A float that is a parameter used to 
                    determine the linear relationship between the 
                    turbulence intensity and the width of the Gaussian 
                    wake shape.
                -   **kb**: A float that is a second parameter used to 
                    determine the linear relationship between the 
                    turbulence intensity and the width of the Gaussian 
                    wake shape.
                -   **alpha**: A float that is a parameter that 
                    determines the dependence of the downstream 
                    boundary between the near wake and far wake region 
                    on the turbulence intensity.
                -   **beta**: A float that is a parameter that 
                    determines the dependence of the downstream 
                    boundary between the near wake and far wake region 
                    on the turbine's induction factor.

    Returns:
        An instantiated Gauss object.
    """

    def __init__(self, parameter_dictionary):
        super().__init__(parameter_dictionary)
        self.model_string = "gauss"
        model_dictionary = parameter_dictionary[self.model_string]
        # wake expansion parameter
        self.ka = float(model_dictionary["ka"])
        # wake expansion parameter
        self.kb = float(model_dictionary["kb"])
        self.alpha = float(model_dictionary["alpha"])  # near wake parameter
        self.beta = float(model_dictionary["beta"])    # near wake parameter

    def function(self, x_locations, y_locations, z_locations, turbine, turbine_coord, deflection_field, wake, flow_field):
        """
        Using the Gaussian wake model, this method calculates and 
        returns the wake velocity deficits, caused by the specified 
        turbine, relative to the freestream velocities at the grid of 
        points comprising the wind farm flow field.

        Args:
            x_locations: An array of floats that contains the 
                streamwise direction grid coordinates of the flow field 
                domain (m).
            y_locations: An array of floats that contains the grid 
                coordinates of the flow field domain in the direction 
                normal to x and parallel to the ground (m).
            z_locations: An array of floats that contains the grid 
                coordinates of the flow field domain in the vertical 
                direction (m).
            turbine: A :py:obj:`floris.simulation.turbine` object that 
                represents the turbine creating the wake.
            turbine_coord: A :py:obj:`floris.utilities.Vec3` object 
                containing the coordinate of the turbine creating the 
                wake (m).
            deflection_field: An array of floats that contains the 
                amount of wake deflection in meters in the y direction 
                at each grid point of the flow field.
            wake: A :py:obj:`floris.simulation.wake` object containing 
                the wake model used.
            flow_field: A :py:class:`floris.simulation.flow_field` 
                object containing the flow field information for the 
                wind farm.

        Returns:
            Three arrays of floats that contain the wake velocity 
            deficit in m/s created by the turbine relative to the 
            freestream velocities for the u, v, and w components, 
            aligned with the x, y, and z directions, respectively. The 
            three arrays contain the velocity deficits at each grid 
            point in the flow field. 
        """

        # veer (degrees)
        veer = flow_field.wind_veer

        # added turbulence model
        TI = turbine.turbulence_intensity

        # turbine parameters
        D = turbine.rotor_diameter
        HH = turbine.hub_height
        yaw = -1 * turbine.yaw_angle  # opposite sign convention in this model
        Ct = turbine.Ct
        U_local = flow_field.u_initial

        # wake deflection
        delta = deflection_field

        # initial velocity deficits
        uR = U_local * Ct / (2.0 * (1 - np.sqrt(1 - (Ct))))
        u0 = U_local * np.sqrt(1 - Ct)

        # initial Gaussian wake expansion
        sigma_z0 = D * 0.5 * np.sqrt(uR / (U_local + u0))
        sigma_y0 = sigma_z0 * cosd(yaw) * cosd(veer)

        # quantity that determines when the far wake starts
        x0 = D * (cosd(yaw) * (1 + np.sqrt(1 - Ct))) / (np.sqrt(2) \
            * (4 * self.alpha * TI + 2 * self.beta * (1 - np.sqrt(1 - Ct)))) \
            + turbine_coord.x1

        # wake expansion parameters
        ky = self.ka * TI + self.kb
        kz = self.ka * TI + self.kb

        # compute velocity deficit
        yR = y_locations - turbine_coord.x2
        xR = yR * tand(yaw) + turbine_coord.x1

        # velocity deficit in the near wake
        sigma_y = (((x0 - xR) - (x_locations - xR)) / (x0 - xR)) * 0.501 * \
            D * np.sqrt(Ct / 2.) + ((x_locations - xR) / (x0 - xR)) * sigma_y0
        sigma_z = (((x0 - xR) - (x_locations - xR)) / (x0 - xR)) * 0.501 * \
            D * np.sqrt(Ct / 2.) + ((x_locations - xR) / (x0 - xR)) * sigma_z0

        sigma_y[x_locations < xR] = 0.5 * D
        sigma_z[x_locations < xR] = 0.5 * D

        a = (cosd(veer)**2) / (2 * sigma_y**2) + \
            (sind(veer)**2) / (2 * sigma_z**2)
        b = -(sind(2 * veer)) / (4 * sigma_y**2) + \
            (sind(2 * veer)) / (4 * sigma_z**2)
        c = (sind(veer)**2) / (2 * sigma_y**2) + \
            (cosd(veer)**2) / (2 * sigma_z**2)
        totGauss = np.exp(-(a * ((y_locations - turbine_coord.x2) - delta)**2 \
                - 2 * b * ((y_locations - turbine_coord.x2) - delta) \
                * ((z_locations - HH)) + c * ((z_locations - HH))**2))

        velDef = (U_local * (1 - np.sqrt(1 - ((Ct * cosd(yaw)) \
                / (8.0 * sigma_y * sigma_z / D**2)))) * totGauss)
        velDef[x_locations < xR] = 0
        velDef[x_locations > x0] = 0

        # wake expansion in the lateral (y) and the vertical (z)
        sigma_y = ky * (x_locations - x0) + sigma_y0
        sigma_z = kz * (x_locations - x0) + sigma_z0

        sigma_y[x_locations < x0] = sigma_y0[x_locations < x0]
        sigma_z[x_locations < x0] = sigma_z0[x_locations < x0]

        # velocity deficit outside the near wake
        a = (cosd(veer)**2) / (2 * sigma_y**2) + \
            (sind(veer)**2) / (2 * sigma_z**2)
        b = -(sind(2 * veer)) / (4 * sigma_y**2) + \
            (sind(2 * veer)) / (4 * sigma_z**2)
        c = (sind(veer)**2) / (2 * sigma_y**2) + \
            (cosd(veer)**2) / (2 * sigma_z**2)
        totGauss = np.exp(-(a * ((y_locations - turbine_coord.x2) - delta)**2 \
                - 2 * b * ((y_locations - turbine_coord.x2) - delta) \
                * ((z_locations - HH)) + c * ((z_locations - HH))**2))

        # compute velocities in the far wake
        velDef1 = (U_local * (1 - np.sqrt(1 - ((Ct * cosd(yaw)) \
                / (8.0 * sigma_y * sigma_z / D**2)))) * totGauss)
        velDef1[x_locations < x0] = 0

        return np.sqrt(velDef**2 + velDef1**2), np.zeros(np.shape(velDef)), np.zeros(np.shape(velDef))


class Curl(WakeVelocity):
    """
    Curl is a wake velocity subclass that contains objects related to 
    the Curled Wake model.

    Curl is a subclass of 
    :py:class:`floris.simulation.wake_velocity.WakeVelocity` that is 
    used to compute the wake velocity deficit based on the curled wake 
    model developed by Martinez-Tossas et al. The curled wake model 
    includes the change in the shape of the wake profile under yawed 
    conditions due to vortices that are shed from the rotor plane of a 
    yawed turbine. The model includes the impact of turbulence 
    intensity, wind veer, and the tip-speed ratio of the turbine. For 
    more information about the curled wake model theory, see: 
    Martinez-Tossas, L. A. et al. "The aerodynamics of the curled wake: 
    a simplified model in view of flow control." *Wind Energy Science*, 
    2019. 

    For more information about the impact of the curled wake behavior 
    on wake steering, see: Fleming, P. et al. "A simulation study 
    demonstreating the importance of large-scale trailing vortices in 
    wake steering." *Wind Energy Science*, 2018. 

    Args:
        parameter_dictionary: A dictionary as generated from the 
            input_reader; it should have the following key-value pairs:

            -   **turbulence_intensity**: A dictionary containing the 
                following key-value pairs used to calculate wake-added 
                turbulence intensity from an upstream turbine, using 
                the approach of Crespo, A. and Herna, J. "Turbulence 
                characteristics in wind-turbine wakes." *J. Wind Eng 
                Ind Aerodyn*. 1996.:

                -   **initial**: A float that is the initial ambient 
                    turbulence intensity, expressed as a decimal 
                    fraction.
                -   **constant**: A float that is the constant used to 
                    scale the wake-added turbulence intensity.
                -   **ai**: A float that is the axial induction factor 
                    exponent used in in the calculation of wake-added 
                    turbulence.
                -   **downstream**: A float that is the exponent 
                    applied to the distance downtream of an upstream 
                    turbine normalized by the rotor diameter used in 
                    the calculation of wake-added turbulence.

            -   **curl**: A dictionary containing the following 
                key-value pairs:

                -   **model_grid_resolution**: A list of three floats 
                    that define the flow field grid resolution in the x,
                    y, and z directions used for the curl wake model 
                    calculations. The grid resolution is specified as 
                    the number of grid points in the flow field domain 
                    in the x, y, and z directions. 
                -   **initial_deficit**: A float that, along with the 
                    freestream velocity and the turbine's induction 
                    factor, is used to determine the initial wake 
                    velocity deficit immediately downstream of the 
                    rotor.
                -   **dissipation**: A float that is a scaling 
                    parameter that determines the amount of dissipation 
                    of the vortices with downstream distance.
                -   **veer_linear**: A float that describes the amount 
                    of linear wind veer. This parameter defines the 
                    linear change in the V velocity between the ground 
                    and hub height, and therefore determines the slope 
                    of the change in the V velocity with height. 

    Returns:
        An instantiated Curl object.
    """

    def __init__(self, parameter_dictionary):
        super().__init__(parameter_dictionary)
        self.model_string = "curl"
        model_dictionary = parameter_dictionary[self.model_string]
        self.model_grid_resolution = Vec3(
            model_dictionary["model_grid_resolution"])
        self.initial_deficit = float(model_dictionary["initial_deficit"])
        self.dissipation = float(model_dictionary["dissipation"])
        self.veer_linear = float(model_dictionary["veer_linear"])
        self.requires_resolution = True

    def function(self, x_locations, y_locations, z_locations, turbine, turbine_coord, deflection_field, wake, flow_field):
        """
        Using the Curl wake model, this method calculates and returns 
        the wake velocity deficits, caused by the specified turbine, 
        relative to the freestream velocities at the grid of points 
        comprising the wind farm flow field.

        Args:
            x_locations: An array of floats that contains the 
                streamwise direction grid coordinates of the flow field 
                domain (m).
            y_locations: An array of floats that contains the grid 
                coordinates of the flow field domain in the direction 
                normal to x and parallel to the ground (m).
            z_locations: An array of floats that contains the grid 
                coordinates of the flow field domain in the vertical 
                direction (m).
            turbine: A :py:obj:`floris.simulation.turbine` object that 
                represents the turbine creating the wake.
            turbine_coord: A :py:obj:`floris.utilities.Vec3` object 
                containing the coordinate of the turbine creating the 
                wake (m).
            deflection_field: An array of floats that contains the 
                amount of wake deflection in meters in the y direction 
                at each grid point of the flow field.
            wake: A :py:obj:`floris.simulation.wake` object containing 
                the wake model used.
            flow_field: A :py:class:`floris.simulation.flow_field` 
                object containing the flow field information for the 
                wind farm.

        Returns:
            Three arrays of floats that contain the wake velocity 
            deficit in m/s created by the turbine relative to the 
            freestream velocities for the u, v, and w components, 
            aligned with the x, y, and z directions, respectively. The 
            three arrays contain the velocity deficits at each grid 
            point in the flow field. 
        """

        # parameters available for tuning to match high-fidelity data
        # parameter for defining initial velocity deficity in the flow field at a turbine
        intial_deficit = self.initial_deficit
        # scaling parameter that adjusts the amount of dissipation of the vortexes
        dissipation = self.dissipation
        # parameter that defines the wind velocity of veer at 0 meters height
        veer_linear = self.veer_linear

        # setup x and y grid information
        x = np.linspace(np.min(x_locations), np.max(
            x_locations), int(self.model_grid_resolution.x1))
        y = np.linspace(np.min(y_locations), np.max(
            y_locations), int(self.model_grid_resolution.x2))

        # find the x-grid location closest to the current turbine
        idx = np.min(np.where(x >= turbine_coord.x1))

        # initialize the flow field
        uw = np.zeros(
            (
                int(self.model_grid_resolution.x1),
                int(self.model_grid_resolution.x2),
                int(self.model_grid_resolution.x3)
            )
        )

        # determine values to create a rotor mask for velocities
        y1 = y_locations[idx, :, :] - turbine_coord.x2
        z1 = z_locations[idx, :, :] - turbine.hub_height
        r1 = np.sqrt(y1**2 + z1**2)

        # add initial velocity deficit at the rotor to the flow field
        uw_initial = -1 * (flow_field.wind_speed * intial_deficit * turbine.aI)
        uw[idx, :, :] = gaussian_filter(
            uw_initial * (r1 <= turbine.rotor_diameter / 2), sigma=1)

        # enforce the boundary conditions
        uw[idx,  0, :] = 0.0
        uw[idx, :,  0] = 0.0
        uw[idx, -1, :] = 0.0
        uw[idx, :, -1] = 0.0

        # TODO: explain?
        uw = -1 * uw

        # parameters to simplify the code
        # diameter of the turbine rotor from the input file
        D = turbine.rotor_diameter
        Ct = turbine.Ct                                 # thrust coefficient of the turbine
        yaw = turbine.yaw_angle                         # yaw angle of the turbine
        HH = turbine.hub_height                         # hub height of the turbine
        # the free-stream velocity of the flow field
        Uinf = flow_field.wind_speed
        # the tip-speed ratior of the turbine
        TSR = turbine.tsr
        # the axial induction factor of the turbine
        aI = turbine.aI
        # initial velocities in the stream-wise, span-wise, and vertical direction
        U, V, W = flow_field.u, flow_field.v, flow_field.w
        # the tilt angle of the rotor of the turbine
        tilt = turbine.tilt_angle

        # calculate the curled wake effects due to the yaw and tilt of the turbine
        Gamma_Yaw = flow_field.air_density * np.pi * D / 8 * Ct * \
            turbine.average_velocity * sind(yaw)
        if turbine.yaw_angle != 0.0:
            YawFlag = 1
        else:
            YawFlag = 0
        Gamma_Tilt = flow_field.air_density * np.pi * D / 8 * Ct * \
            turbine.average_velocity * sind(tilt)
        if turbine.tilt_angle != 0.0:
            TiltFlag = 1
        else:
            TiltFlag = 0

        Gamma = Gamma_Yaw + Gamma_Tilt

        # calculate the curled wake effects due to the rotation of the turbine rotor
        Gamma_wake_rotation = 2 * np.pi * D * (aI - aI**2) * Uinf / TSR

        # =======================================================================
        # add curl Elliptic
        # =======================================================================
        eps = 0.2 * D

        # distribute rotation across the blade
        z_vector = np.linspace(0, D / 2, 100)

        # length of each section dz
        dz = z_vector[1] - z_vector[0]

        # scale the circulation of each section dz
        if yaw != 0 or tilt != 0:
            Gamma0 = 4 / np.pi * Gamma
        else:
            Gamma0 = 0.0

        # loop through all the vortices from an elliptic wind distribution
        # skip the last point because it has zero circulation
        for z in z_vector[:-1]:

            # Compute the non-dimensional circulation
            Gamma = (-4 * Gamma0 * z * dz /
                     (D**2 * np.sqrt(1 - (2 * z / D)**2)))

            # locations of the tip vortices
            # top
            y_vortex_1 = turbine_coord.x2 + z * TiltFlag
            z_vortex_1 = HH + z * YawFlag

            # bottom
            y_vortex_2 = turbine_coord.x2 - z * TiltFlag
            z_vortex_2 = HH - z * YawFlag

            # vortex velocities
            # top
            v1, w1 = self._vortex(flow_field.y[idx, :, :] - y_vortex_1, flow_field.z[idx, :, :]
                                  - z_vortex_1, flow_field.x[idx, :, :] - turbine_coord.x1, -Gamma, eps, Uinf)
            # bottom
            v2, w2 = self._vortex(flow_field.y[idx, :, :] - y_vortex_2, flow_field.z[idx, :, :]
                                  - z_vortex_2, flow_field.x[idx, :, :] - turbine_coord.x1, Gamma, eps, Uinf)

            # add ground effects
            v3, w3 = self._vortex(flow_field.y[idx, :, :] - y_vortex_1, flow_field.z[idx, :, :]
                                  + z_vortex_1, flow_field.x[idx, :, :] - turbine_coord.x1, Gamma, eps, Uinf)
            v4, w4 = self._vortex(flow_field.y[idx, :, :] - y_vortex_2, flow_field.z[idx, :, :]
                                  + z_vortex_2, flow_field.x[idx, :, :] - turbine_coord.x1, -Gamma, eps, Uinf)

            V[idx, :, :] += v1 + v2 + v3 + v4
            W[idx, :, :] += w1 + w2 + w3 + w4

        # add wake rotation
        v5, w5 = self._vortex(flow_field.y[idx, :, :] - turbine_coord.x2, flow_field.z[idx, :, :]
                              - turbine.hub_height, flow_field.x[idx, :, :]
                              - turbine_coord.x1, Gamma_wake_rotation, 0.2 * D, Uinf) \
            * (np.sqrt((flow_field.y[idx, :, :] - turbine_coord.x2)**2
                       + (flow_field.z[idx, :, :] - turbine.hub_height)**2) <= D/2)
        v6, w6 = self._vortex(flow_field.y[idx, :, :] - turbine_coord.x2, flow_field.z[idx, :, :]
                              + turbine.hub_height, flow_field.x[idx, :, :]
                              - turbine_coord.x1, -Gamma_wake_rotation, 0.2 * D, Uinf) \
            * (np.sqrt((flow_field.y[idx, :, :] - turbine_coord.x2)**2
                       + (flow_field.z[idx, :, :] - turbine.hub_height)**2) <= D/2)
        V[idx, :, :] += v5 + v6
        W[idx, :, :] += w5 + w6

        # decay the vortices as they move downstream
        lmda = 15
        kappa = 0.41
        lm = kappa * z / (1 + kappa * z / lmda)
        dudz_initial = np.gradient(U, z, axis=2)
        nu = lm**2 * np.abs(dudz_initial[0, :, :])

        for i in range(idx, len(x) - 1):
            V[i + 1, :, :] = V[idx, :, :] * eps**2 \
                / (4 * nu * (flow_field.x[i, :, :]
                             - turbine_coord.x1) / Uinf + eps**2)
            W[i + 1, :, :] = W[idx, :, :] * eps**2 \
                / (4 * nu * (flow_field.x[i, :, :]
                             - turbine_coord.x1) / Uinf + eps**2)

        # simple implementation of linear veer, added to the V component of the flow field
        z = np.linspace(
            np.min(z_locations),
            np.max(z_locations),
            int(self.model_grid_resolution.x3)
        )
        z_min = HH
        b_veer = veer_linear
        m_veer = -b_veer / z_min

        v_veer = m_veer * z + b_veer

        for i in range(len(z) - 1):
            V[:, :, i] = V[:, :, i] + v_veer[i]

        # ===========================================================================================
        # SOLVE CURL
        # ===========================================================================================
        dudz_initial = np.gradient(U, axis=2) \
            / np.gradient(z_locations, axis=2)

        ti_initial = flow_field.turbulence_intensity

        # turbulence intensity parameters stored in floris.json
        ti_i = self.ti_initial
        ti_constant = self.ti_constant
        ti_ai = self.ti_ai
        ti_downstream = self.ti_downstream

        for i in range(idx + 1, len(x)):

            # compute the change in x
            dx = x[i] - x[i - 1]

            dudy = np.gradient(uw[i - 1, :, :], axis=0) \
                / np.gradient(y_locations[i - 1, :, :], axis=0)
            dudz = np.gradient(uw[i - 1, :, :], axis=1) \
                / np.gradient(z_locations[i - 1, :, :], axis=1)

            gradU = np.gradient(np.gradient(uw[i - 1, :, :], axis=0), axis=0) \
                / np.gradient(y_locations[i - 1, :, :], axis=0)**2 \
                + np.gradient(np.gradient(uw[i - 1, :, :], axis=1), axis=1) \
                / np.gradient(z_locations[i - 1, :, :], axis=1)**2

            lm = kappa * z / (1 + kappa * z / lmda)
            nu = lm**2 * np.abs(dudz_initial[i - 1, :, :])

            # turbulence intensity calculation based on Crespo et. al.
            ti_local = 10*ti_constant \
                * turbine.aI**ti_ai \
                * ti_initial**ti_i \
                * ((x[i] - turbine_coord.x1) / turbine.rotor_diameter)**ti_downstream

            # solve the marching problem for u, v, and w
            uw[i, :, :] = uw[i - 1, :, :] + (dx / (U[i - 1, :, :])) \
                * (-V[i - 1, :, :] * dudy - W[i - 1, :, :] * dudz
                   + dissipation * D * nu * ti_local * gradU)
            # enforce boundary conditions
            uw[i, :, 0] = np.zeros(len(y))
            uw[i, 0, :] = np.zeros(len(z))

        uw[x_locations < turbine_coord.x1] = 0.0

        return uw, V, W

    def _vortex(self, x, y, z, Gamma, eps, U):
        # compute the vortex velocity
        v = (Gamma / (2 * np.pi)) * (y / (x**2 + y**2)) \
            * (1 - np.exp(-(x**2 + y**2) / eps**2))
        w = -(Gamma / (2 * np.pi)) * (x / (x**2 + y**2)) \
            * (1 - np.exp(-(x**2 + y**2) / eps**2))

        return v, w
